# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import inspect
import os
from types import FunctionType, MethodType
from typing import Any, Callable, Iterable, Optional, TypeVar

#: :meta private:
Method = TypeVar("Method", bound=Callable[..., Any])


#  pylint: disable=invalid-name
class overrides(object):
    """
    A decorator for indicating that a method overrides another method in
    a superclass.  This checks that the method does actually exist,
    copies the doc-string for the method, and enforces that the method
    overridden is specified, making maintenance easier.
    """
    # This near constant is changed by unit tests to check our code
    # Github actions sets TYPE_OVERRIDES as True
    __CHECK_TYPES = os.getenv("TYPE_OVERRIDES")

    __slots__ = [
        # The method in the superclass that this method overrides
        "_superclass_method",
        # True if the doc string is to be extended, False to set if not set
        "_extend_doc",
        # Any additional arguments required by the subclass method
        "_additional_arguments",
        # True if the subclass method may have additional defaults
        "_extend_defaults",
        # True if the name check is relaxed
        "_relax_name_check",
        # The name of the thing being overridden for error messages
        "_override_name",
        # True if this method adds typing info not in the original
        "_adds_typing"
    ]

    def __init__(
            self, super_class_method, *, extend_doc: bool = True,
            additional_arguments: Optional[Iterable[str]] = None,
            extend_defaults: bool = False, adds_typing: bool = False,):
        """
        :param super_class_method: The method to override in the superclass
        :param bool extend_doc:
            True the method doc string should be appended to the super-method
            doc string, False if the documentation should be set to the
            super-method doc string only if there isn't a doc string already
        :param iterable(str) additional_arguments:
            Additional arguments taken by the subclass method over the
            superclass method, e.g., that are to be injected
        :param bool extend_defaults:
            Whether the subclass may specify extra defaults for the parameters
        :param adds_typing:
            Allows more typing (of non additional) than in the subclass.
            Should only be used for built in super classes
        """
        if isinstance(super_class_method, property):
            super_class_method = super_class_method.fget
        if not isinstance(super_class_method, (FunctionType, MethodType)):
            raise TypeError("may only decorate method declarations; "
                            f"this is a {type(super_class_method)}")
        self._superclass_method = super_class_method
        self._extend_doc = bool(extend_doc)
        self._extend_defaults = bool(extend_defaults)
        self._relax_name_check = False
        self._override_name = "super class method"
        if additional_arguments is not None:
            self._additional_arguments = frozenset(additional_arguments)
        else:
            self._additional_arguments = frozenset()
        self._adds_typing = adds_typing

    @staticmethod
    def __match_defaults(default_args, super_defaults, extend_ok):
        if default_args is None:
            return super_defaults is None
        elif super_defaults is None:
            return extend_ok
        if extend_ok:
            return len(default_args) >= len(super_defaults)
        return len(default_args) == len(super_defaults)

    def _verify_types(self, method_args, super_args, all_args):
        """
        Check that the arguments match.
        """
        if not self.__CHECK_TYPES:
            return
        if "self" in all_args:
            all_args.remove("self")
        elif "cls" in all_args:
            all_args.remove("cls")
        method_types = method_args.annotations
        super_types = super_args.annotations
        for arg in all_args:
            if arg not in super_types and not self._adds_typing:
                raise AttributeError(
                    f"Super Method {self._superclass_method.__name__} "
                    f"has untyped arguments including {arg}")
            if arg not in method_types:
                raise AttributeError(
                    f"Method {self._superclass_method.__name__} "
                    f"has untyped arguments including {arg}")

        if len(all_args) == 0:
            if "return" not in super_types and not self._adds_typing and \
                    not method_args.varkw and not method_args.varargs:
                raise AttributeError(
                    f"Super Method {self._superclass_method.__name__} "
                    f"has no arguments so should declare a return type")
            if "return" not in method_types and \
                    not method_args.varkw and not method_args.varargs:
                raise AttributeError(
                    f"Method {self._superclass_method.__name__} "
                    f"has no arguments so should declare a return type")

        if "return" in super_types:
            if "return" not in method_types:
                if super_types["return"] is not None:
                    raise AttributeError(
                        f"Method {self._superclass_method.__name__} "
                        f"has no return type, while super does")
        else:
            if "return" in method_types and not self._adds_typing:
                if method_types["return"] is not None:
                    raise AttributeError(
                        f"Super Method {self._superclass_method.__name__} "
                        f"has no return type, while this does")

    def __verify_method_arguments(self, method: Method):
        """
        Check that the arguments match.
        """
        method_args = inspect.getfullargspec(method)
        super_args = inspect.getfullargspec(self._superclass_method)
        all_args = [
            arg for arg in method_args.args
            if arg not in self._additional_arguments]
        default_args = None
        if method_args.defaults is not None:
            default_args = [
                arg for arg in method_args.defaults
                if arg not in self._additional_arguments]
        if len(all_args) != len(super_args.args):
            raise AttributeError(
                f"Method has {len(method_args.args)} arguments but "
                f"{self._override_name} has {len(super_args.args)} arguments")
        for arg, super_arg in zip(all_args, super_args.args):
            if arg != super_arg:
                raise AttributeError(f"Missing argument {super_arg}")
        if not self.__match_defaults(
                default_args, super_args.defaults, self._extend_defaults):
            raise AttributeError(
                f"Default arguments don't match {self._override_name}")
        self._verify_types(method_args, super_args, all_args)

    def __call__(self, method: Method) -> Method:
        """
        Apply the decorator to the given method.
        """
        # Check and fail if this is a property
        if isinstance(method, property):
            raise AttributeError(
                f"Please ensure that the {self.__class__.__name__} decorator "
                "is the last decorator before the method declaration")

        # Check that the name matches
        if (not self._relax_name_check and
                method.__name__ != self._superclass_method.__name__):
            raise AttributeError(
                f"{self._override_name} name "
                f"{self._superclass_method.__name__} does not match "
                f"{method.__name__}. Ensure {self.__class__.__name__} is the "
                "last decorator before the method declaration")

        # Check that the arguments match (except for __init__ as this might
        # take extra arguments or pass arguments not specified)
        if method.__name__ != "__init__":
            self.__verify_method_arguments(method)

        if (self._superclass_method.__doc__ is not None and
                method.__doc__ is None):
            method.__doc__ = self._superclass_method.__doc__
        elif (self._extend_doc and
                self._superclass_method.__doc__ is not None):
            method.__doc__ = (
                self._superclass_method.__doc__ + (method.__doc__ or ""))
        return method

    @classmethod
    def check_types(cls):
        """
        If called will trigger check that all parameters are checked.

        Used for testing, to avoid users being affected by the strict checks
        """
        cls.__CHECK_TYPES = True



# Copyright (c) 2018 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import platform
import subprocess
import time
from typing import Set


class Ping(object):
    """
    Platform-independent ping support.
    """

    #: The unreachable host cache.
    unreachable: Set[str] = set()

    @staticmethod
    def ping(ip_address):
        """
        Send a ping (ICMP ECHO request) to the given host.
        SpiNNaker boards support ICMP ECHO when booted.

        :param str ip_address:
            The IP address to ping. Hostnames can be used, but are not
            recommended.
        :return:
            return code of subprocess; 0 for success, anything else for failure
        :rtype: int
        """
        if platform.platform().lower().startswith("windows"):
            cmd = "ping -n 1 -w 1 "
        else:
            cmd = "ping -c 1 -W 1 "
        process = subprocess.Popen(
            cmd + ip_address, shell=True, stdout=subprocess.PIPE)
        time.sleep(1.2)
        process.stdout.close()
        process.wait()
        return process.returncode

    @staticmethod
    def host_is_reachable(ip_address):
        """
        Test if a host is unreachable via ICMP ECHO.

        .. note::
            This information may be cached in various ways. Transient failures
            are not necessarily detected or recovered from.

        :param str ip_address:
            The IP address to ping. Hostnames can be used, but are not
            recommended.
        :rtype: bool
        """
        if ip_address in Ping.unreachable:
            return False
        tries = 0
        while True:
            if Ping.ping(ip_address) == 0:
                return True
            tries += 1
            if tries > 10:
                Ping.unreachable.add(ip_address)
                return False



# Copyright (c) 2018 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os
import shutil
import struct
import sys
from typing import Optional, Tuple
from spinn_utilities.overrides import overrides
from spinn_utilities.config_holder import get_config_str_or_none
from spinn_utilities.log import FormatAdapter
from .file_converter import FORMAT_EXP
from .file_converter import TOKEN
from .log_sqllite_database import DB_FILE_NAME, LogSqlLiteDatabase

logger = FormatAdapter(logging.getLogger(__name__))

LEVELS = {10: "[DEBUG]",
          20: "[INFO]",
          30: "[WARN]",
          40: "[ERROR]"}


class Replacer(LogSqlLiteDatabase):
    """
    Performs replacements.
    """

    @overrides(LogSqlLiteDatabase._database_file)
    def _database_file(self) -> str:
        database_file = super()._database_file()
        if not os.path.isfile(database_file):
            external_binaries = get_config_str_or_none(
                "Mapping", "external_binaries")
            if external_binaries is not None:
                source_file = os.path.join(external_binaries, DB_FILE_NAME)
                if os.path.exists(source_file):

                    shutil.copyfile(source_file, database_file)
        return database_file

    @overrides(LogSqlLiteDatabase._extra_database_error_message)
    def _extra_database_error_message(self) -> str:
        extra__binaries = get_config_str_or_none(
            "Mapping", "external_binaries")
        if extra__binaries is None:
            return ""
        else:
            return (f"The cfg {extra__binaries=} "
                    f"also does not contain a {DB_FILE_NAME}. ")

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        # nothing yet
        pass

    _INT_FMT = struct.Struct("!I")
    _FLT_FMT = struct.Struct("!f")
    _DBL_FMT = struct.Struct("!d")

    def _replace(self, short: str) -> Optional[Tuple[int, str, int, str]]:
        """
        Apply the replacements to a short message.

        :param str short: The short message to apply the transform to.
        :return: The expanded message.
        :rtype: str
        """
        parts = short.split(TOKEN)
        if not parts[0].isdigit():
            return None
        data = self.get_log_info(parts[0])
        if data is None:
            return None
        (log_level, file_name, line_num, original) = data

        replaced = original.encode("latin-1").decode("unicode_escape")
        if len(parts) > 1:
            matches = FORMAT_EXP.findall(original)
            # Remove any blanks due to double spacing
            matches = [x for x in matches if x != ""]
            # Start at 0 so first i+1 puts you at 1 as part 0 is the short
            i = 0
            try:
                for match in matches:
                    i += 1
                    if match.endswith("f"):
                        replacement = str(self._hex_to_float(parts[i]))
                    elif match.endswith("F"):
                        replacement = str(self._hexes_to_double(
                            parts[i], parts[i+1]))
                        i += 1
                    else:
                        replacement = parts[i]
                    replaced = replaced.replace(match, replacement, 1)
            except Exception:  # pylint: disable=broad-except
                return None
        return (log_level, file_name, line_num, replaced)

    def replace(self, short: str) -> str:
        """
        Apply the replacements to a short message.

        :param str short:
        :rtype: str
        """
        data = self._replace(short)
        if data is None:
            return short
        (log_level, file_name, line_num, replaced) = data
        return f"{LEVELS[log_level]} ({file_name}: {line_num}): {replaced}"

    def _hex_to_float(self, hex_str):
        return self._FLT_FMT.unpack(
            self._INT_FMT.pack(int(hex_str, 16)))[0]

    def _hexes_to_double(self, upper, lower):
        return self._DBL_FMT.unpack(
            self._INT_FMT.pack(int(upper, 16)) +
            self._INT_FMT.pack(int(lower, 16)))[0]


if __name__ == '__main__':
    encoded = sys.argv[1]
    LINE = "".join([c if c.isalnum() else TOKEN for c in encoded])
    with Replacer() as replacer:
        print(replacer.replace(LINE))



# Copyright (c) 2018 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import yaml
import io
import importlib
import argparse
import sys
from .citation_updater_and_doi_generator import CitationUpdaterAndDoiGenerator

ENCODING = "utf-8"

REQUIREMENTS_FILE = "requirements.txt"
C_REQUIREMENTS_FILE = "c_requirements.txt"
CITATION_FILE = "CITATION.cff"
PYPI_TO_IMPORT_FILE = "pypi_to_import"

REFERENCES_YAML_POINTER = "references"
REFERENCE_TYPE = "software"
REFERENCES_AUTHORS_TYPE = "authors"
REFERENCES_TITLE_TYPE = "title"
REFERENCES_VERSION_TYPE = "version"
REFERENCES_DATE_TYPE = "date-released"
REFERENCES_URL_TYPE = "url"
REFERENCES_REPO_TYPE = "repository"
REFERENCES_CONTACT_TYPE = "contact"
REFERENCES_TYPE_TYPE = "type"
REFERENCES_SOFTWARE_TYPE = "software"

CITATION_DOI_TYPE = 'identifier'

# pylint: skip-file


class CitationAggregator(object):
    """
    Helper class for building a citation file which references all
    dependencies.
    """

    def create_aggregated_citation_file(
            self, module_to_start_at, aggregated_citation_file):
        """
        Entrance method for building the aggregated citation file.

        :param module_to_start_at:
            the top level module to figure out its citation file for
        :type module_to_start_at: python module
        :param str aggregated_citation_file:
            file name of aggregated citation file
        """
        # get the top citation file to add references to
        top_citation_file_path = os.path.join(os.path.dirname(os.path.dirname(
            os.path.abspath(module_to_start_at.__file__))), CITATION_FILE)
        modules_seen_so_far = set()
        modules_seen_so_far.add("")  # Make sure the empty entry is absent
        with open(top_citation_file_path, encoding=ENCODING) as stream:
            top_citation_file = yaml.safe_load(stream)
        top_citation_file[REFERENCES_YAML_POINTER] = list()

        # get the dependency list
        requirements_file_path = os.path.join(
            os.path.dirname(os.path.dirname(os.path.abspath(
                module_to_start_at.__file__))), REQUIREMENTS_FILE)
        c_requirements_file_path = os.path.join(
            os.path.dirname(os.path.dirname(os.path.abspath(
                module_to_start_at.__file__))), C_REQUIREMENTS_FILE)

        # attempt to get python PYPI to import command map
        pypi_to_import_map_file = os.path.join(
            os.path.dirname(os.path.dirname(os.path.abspath(
                module_to_start_at.__file__))),
            PYPI_TO_IMPORT_FILE)
        pypi_to_import_map = None
        if os.path.isfile(pypi_to_import_map_file):
            pypi_to_import_map = self._read_pypi_import_map(
                pypi_to_import_map_file)

        if os.path.isfile(requirements_file_path):
            with open(requirements_file_path, encoding="utf-8") as r_file:
                for line in r_file:
                    module = line.strip()
                    if module.startswith("#"):
                        continue
                    if module not in modules_seen_so_far:
                        import_name = pypi_to_import_map.get(module, module)
                        #  pylint: disable=broad-except
                        try:
                            imported_module = importlib.import_module(
                                import_name)
                            self._handle_python_dependency(
                                top_citation_file, imported_module,
                                modules_seen_so_far,
                                pypi_to_import_map[module])
                        except Exception as e:  # pragma: no cover
                            print("Error handling python dependency "
                                  f"{module}: {e}")

        if os.path.isfile(c_requirements_file_path):
            with open(c_requirements_file_path, encoding=ENCODING) as r_file:
                for line in r_file:
                    module = line.strip()
                    if module.startswith("#"):
                        continue
                    if module not in modules_seen_so_far:
                        self._handle_c_dependency(
                            top_citation_file, module, modules_seen_so_far)

        # write citation file with updated fields
        with io.open(
                aggregated_citation_file, 'w', encoding=ENCODING) as outfile:
            yaml.dump(top_citation_file, outfile, default_flow_style=False,
                      allow_unicode=True)

    @staticmethod
    def _read_pypi_import_map(aggregated_citation_file):
        """
        Read the PYPI to import name map.

        :param str aggregated_citation_file: path to the PYPI map file
        :return: map between PYPI names and import names
        :rtype: dict(str,str)
        """
        pypi_to_import_map = dict()
        with open(aggregated_citation_file, encoding=ENCODING) as f:
            for line in f:
                [pypi, import_command] = line.split(":")
                pypi_to_import_map[pypi] = import_command.split("\n")[0]
        return pypi_to_import_map

    def _handle_c_dependency(
            self, top_citation_file, module, modules_seen_so_far):
        """
        Handle a C code dependency.

        :param str top_citation_file: YAML file for the top citation file
        :param str module: module to find
        :param set(str) modules_seen_so_far:
        """
        cleaned_path = self.locate_path_for_c_dependency(module)
        if cleaned_path is not None:
            # process reference
            reference_entry = self._process_reference(
                cleaned_path, None, modules_seen_so_far, module)

            # append to the top citation file
            top_citation_file[REFERENCES_YAML_POINTER].append(
                reference_entry)
            self._search_for_other_c_references(
                reference_entry, cleaned_path, modules_seen_so_far)
        else:
            print(f"Could not find C dependency {module}")

    @staticmethod
    def locate_path_for_c_dependency(true_software_name):
        """
        :param str true_software_name:
        :rtype: str or None
        """
        environment_path_variable = os.environ.get('PATH')
        if environment_path_variable is not None:
            software_paths = environment_path_variable.split(":")
            for software_path in software_paths:
                # clear path to have repository name at end
                last_version = None
                cleaned_path = software_path
                while ((cleaned_path != last_version) and (
                        not (cleaned_path.split(os.sep)[-1] ==
                             true_software_name))):
                    last_version = cleaned_path
                    cleaned_path = os.path.dirname(cleaned_path)

                if cleaned_path != last_version:
                    return cleaned_path
        return None

    def _search_for_other_c_references(
            self, reference_entry, software_path, modules_seen_so_far):
        """
        Go though the top level path and tries to locate other CFF
        files that need to be added to the references pile.

        :param dict(str,list(str)) reference_entry:
            The reference entry to add new dependencies as references for.
        :param str software_path: the path to search in
        :param set(str) modules_seen_so_far:
        """
        for possible_extra_citation_file in os.listdir(software_path):
            if possible_extra_citation_file.endswith(".cff"):
                dependency_reference_entry = \
                    self._read_and_process_reference_entry(
                        os.path.join(software_path,
                                     possible_extra_citation_file))
                reference_entry[REFERENCES_YAML_POINTER] = list()
                reference_entry[REFERENCES_YAML_POINTER].append(
                    dependency_reference_entry)
                modules_seen_so_far.add(
                    possible_extra_citation_file.split(".")[0])

    def _handle_python_dependency(
            self, top_citation_file, imported_module, modules_seen_so_far,
            module_name):
        """
        Handle a python dependency.

        :param dict(str,list(str)) top_citation_file:
            YAML file for the top citation file
        :param imported_module: the actual imported module
        :type imported_module: ModuleType
        :param set(str) modules_seen_so_far:
            list of names of dependencies already processed
        :param str module_name:
            the name of this module to consider as a dependency
        :raises FileNotFoundError:
        """
        # get modules citation file
        citation_level_dir = os.path.abspath(imported_module.__file__)
        m_path = module_name.replace(".", os.sep)
        last_citation_level_dir = None
        while (not citation_level_dir.endswith(m_path) and
               last_citation_level_dir != citation_level_dir):
            last_citation_level_dir = citation_level_dir
            citation_level_dir = os.path.dirname(citation_level_dir)
        if citation_level_dir == last_citation_level_dir:  # pragma: no cover
            raise FileNotFoundError(
                f"Folder for module {module_name} not found")

        # get the reference data for the reference
        reference_entry = self._process_reference(
            citation_level_dir, imported_module, modules_seen_so_far,
            module_name)

        if reference_entry is not None:
            # append to the top citation file
            top_citation_file[REFERENCES_YAML_POINTER].append(reference_entry)

    def _process_reference(
            self, citation_level_dir, imported_module, modules_seen_so_far,
            module_name):
        """
        Take a module level and tries to locate and process a citation file.

        :param str citation_level_dir:
            the expected level where the ``CITATION.cff`` should be
        :param imported_module: the module after being imported
        :type imported_module: python module
        :param set(str) modules_seen_so_far:
            list of dependencies already processed
        :return: the reference entry in JSON format
        :rtype: dict
        """
        # if it exists, add it as a reference to the top one
        if os.path.isfile(os.path.join(citation_level_dir, CITATION_FILE)):
            reference_entry = self._read_and_process_reference_entry(
                os.path.join(citation_level_dir, CITATION_FILE))

        # check that the file isn't one above (not installed, but developer
        # mode)
        elif os.path.isfile(os.path.join(os.path.dirname(
                os.path.abspath(citation_level_dir)), CITATION_FILE)):
            reference_entry = self._read_and_process_reference_entry(
                os.path.join(os.path.dirname(
                    os.path.abspath(citation_level_dir)), CITATION_FILE))

        # if no citation file exists, do an attempt to find a version to build
        # from
        else:
            # one from version
            reference_entry = self._try_to_find_version(
                imported_module, module_name)

        modules_seen_so_far.add(imported_module)
        return reference_entry

    @staticmethod
    def _try_to_find_version(imported_module, module_name):
        """
        Try to locate a version file or version data to auto-generate
        minimal citation data.

        :param imported_module:
            the module currently trying to find the version of
        :type imported_module: python module
        :return: reference entry for this python module
        :rtype: dict
        """
        reference_entry = dict()
        reference_entry[REFERENCES_TYPE_TYPE] = REFERENCES_SOFTWARE_TYPE
        reference_entry[REFERENCES_TITLE_TYPE] = module_name
        if (hasattr(imported_module, "__version_day__") and
                hasattr(imported_module, "__version_month__") and
                hasattr(imported_module, "__version_year__")):
            reference_entry[REFERENCES_DATE_TYPE] = \
                CitationUpdaterAndDoiGenerator.\
                convert_text_date_to_date(
                    version_day=imported_module.__version_day__,
                    version_month=imported_module.__version_month__,
                    version_year=imported_module.__version_year__)
        if hasattr(imported_module, "__version__"):
            reference_entry[REFERENCES_VERSION_TYPE] = \
                imported_module.__version__
        elif hasattr(imported_module, "version"):
            reference_entry[REFERENCES_VERSION_TYPE] = \
                imported_module.version
        elif hasattr(imported_module, "_version"):
            # pylint: disable=protected-access
            reference_entry[REFERENCES_VERSION_TYPE] = \
                imported_module._version
        return reference_entry

    @staticmethod
    def _read_and_process_reference_entry(dependency_citation_file_path):
        """
        Read a ``CITATION.cff`` and makes it a reference for a higher
        level citation file.

        :param str dependency_citation_file_path:
            path to a `CITATION.cff` file
        :return: reference entry for the higher level `CITATION.cff`
        :rtype: dict
        """
        reference_entry = dict()

        with open(dependency_citation_file_path, 'r', encoding="utf-8") \
                as stream:
            dependency_citation_file = yaml.safe_load(stream)

            reference_entry[REFERENCES_TYPE_TYPE] = REFERENCES_SOFTWARE_TYPE
            reference_entry[REFERENCES_AUTHORS_TYPE] = \
                dependency_citation_file[REFERENCES_AUTHORS_TYPE]
            reference_entry[REFERENCES_TITLE_TYPE] = \
                dependency_citation_file[REFERENCES_TITLE_TYPE]
            reference_entry[REFERENCES_CONTACT_TYPE] = \
                dependency_citation_file[REFERENCES_CONTACT_TYPE]
            reference_entry[REFERENCES_VERSION_TYPE] = \
                dependency_citation_file[REFERENCES_VERSION_TYPE]
            reference_entry[REFERENCES_DATE_TYPE] = \
                dependency_citation_file[REFERENCES_DATE_TYPE]
            reference_entry[REFERENCES_URL_TYPE] = \
                dependency_citation_file[REFERENCES_URL_TYPE]
            reference_entry[REFERENCES_REPO_TYPE] = \
                dependency_citation_file[REFERENCES_REPO_TYPE]
        return reference_entry


def generate_aggregate(arguments=None):
    """
    Command-line tool to generate a single ``citation.cff`` from others.

    :param list(str) arguments: Command line arguments.

        * ``--output_path``: \
            Where to write the aggregate file
        * ``--top_module``: \
            The module to start aggregating the ``citation.cff``\\s from
        * ``--doi_title``: \
            The title of the DOI
        * ``--zenodo_access_token``: \
            The access token for Zenodo
        * ``--tools_doi``: \
            The DOI of the tools
    """
    parser = argparse.ArgumentParser(description="Aggregate Citations")
    parser.add_argument("output_path", help="The file to store the result in")
    parser.add_argument("top_module", help="The module to start with")
    parser.add_argument("--create_doi", action="store_true",
                        help="Create a DOI from the resulting citation"
                             " on Zenodo")
    parser.add_argument("--publish_doi", action="store_true",
                        help="Publish the DOI created")
    parser.add_argument("--doi_title",
                        help="The title to give the created DOI")
    parser.add_argument("--previous_doi",
                        help="The DOI this is a newer version of")
    parser.add_argument("--zenodo_access_token",
                        help="Access token for Zenodo")

    args = parser.parse_args(arguments)
    error = False
    if args.create_doi:  # pragma: no cover
        if not args.doi_title:
            print("--doi_title required when creating a DOI")
            error = True
        if not args.previous_doi:
            print("--previous_doi required when creating a DOI")
            error = True
        if not args.zenodo_access_token:
            print("--zenodo_access_token required when creating a DOI")
            error = True
    if args.publish_doi and not args.create_doi:  # pragma: no cover
        print("Cannot publish DOI without creating one")
        error = True
    if error:  # pragma: no cover
        parser.print_usage()
        sys.exit()

    top_module = importlib.import_module(args.top_module)

    citation_aggregator = CitationAggregator()
    citation_aggregator.create_aggregated_citation_file(
        top_module, args.output_path)
    citation_updater_and_dio_generator = CitationUpdaterAndDoiGenerator()
    citation_updater_and_dio_generator.update_citation_file_and_create_doi(
        citation_file_path=args.output_path,
        doi_title=args.doi_title,
        create_doi=args.create_doi, publish_doi=args.publish_doi,
        previous_doi=args.previous_doi,
        zenodo_access_token=args.zenodo_access_token,
        module_path=os.path.dirname(top_module.__path__[0]))


if __name__ == "__main__":
    generate_aggregate()  # pragma: no cover



# Copyright (c) 2021 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from enum import Enum


class RunStatus(Enum):
    """
    Different states the run could be in.

    This is from the perspective of the users script.
    It says nothing about if there is C code running,
    i.e., has a `sim.run` (or similar) call started but not yet returned.

    This is combined with the :py:class:`ResetStatus` to get the needed status.

    This class is design to used internally by :py:class:`UtilsDataView`.
    """
    NOT_SETUP = 0
    NOT_RUNNING = 1
    IN_RUN = 2
    STOP_REQUESTED = 3
    STOPPING = 4
    SHUTDOWN = 5



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class MultipleValuesException(Exception):
    """
    Raised when there more than one value found unexpectedly.
    """

    def __init__(self, key=None, value1=None, value2=None):
        if key is None:
            msg = "Multiple values found"
        else:
            msg = f"Multiple values found for key {key}"
        if value1 is not None and value2 is not None:
            msg += f" values found include {value1} and {value2}"
        super().__init__(msg)



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations
from typing import Generic, Iterable, Optional, Union, TYPE_CHECKING
from spinn_utilities.overrides import overrides
from .abstract_dict import AbstractDict, T
if TYPE_CHECKING:
    from .range_dictionary import RangeDictionary


class AbstractView(  # pylint: disable=abstract-method
        AbstractDict[T], Generic[T]):
    """
    A view over a ranged dictionary.

    .. note::
        The view may currently be read from only with int and int-collection
        indices, and only be written to with str indices. This may change to
        become more permissive in future versions.
    """
    __slots__ = ["_range_dict"]

    def __init__(self, range_dict: RangeDictionary[T]):
        """
        Use :py:meth:`RangeDictionary.view_factory` to create views
        """
        self._range_dict = range_dict

    def __getitem__(self, key: Union[int, slice, Iterable[int]]):
        """
        Support for the view[x] based the type of the key

        key is a str is currently not supported use get_value instead.
        In the future this may be supported to return some kind of list
        (AbstractList) but how to handle a view there to be determined

        For int and int collections a new view will be returned using
        :py:meth:`RangeDictionary.view_factory`

        .. note::
            int are indexes into the list of IDs not ID values.
            So for a view with IDs [2,3,4,5] view[2] will have an ID of 4

        :param key: str, int or collection of int
        :return: the single value for the str key or a view
        :raises MultipleValuesException: If the keys has multiple values set.
            But not if other keys not asked for have multiple values
        """
        if isinstance(key, str):
            raise KeyError("view[key] is not supported Use get_value() ")
        ids = self.ids()
        if isinstance(key, (slice, int)):
            return self._range_dict.view_factory(ids[key])
        return self._range_dict.view_factory([ids[i] for i in key])

    def __setitem__(self, key: str, value: T):
        """
        See :py:meth:`AbstractDict.set_value`

        .. note::
            Unlike ``__getitem__``, int based IDs are *not* supported so
            ``view[int] ==`` will raise an exception
        """
        if isinstance(key, str):
            return self.set_value(key=key, value=value)
        if isinstance(key, (slice, int, tuple, list)):
            raise KeyError("Setting of a slice/IDs not supported")
        raise KeyError(f"Unexpected key type: {type(key)}")

    @overrides(AbstractDict.get_default)
    def get_default(self, key: str) -> Optional[T]:
        return self._range_dict.get_default(key)

    @overrides(AbstractDict.keys)
    def keys(self) -> Iterable[str]:
        return self._range_dict.keys()



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations
from numbers import Number
from typing import (
    Any, Callable, Generic, Iterator, Optional, Sequence, Tuple,
    TypeVar, Union, cast)
import numpy
from numpy.typing import NDArray
from typing_extensions import Self, TypeAlias, TypeGuard
from spinn_utilities.abstract_base import AbstractBase, abstractmethod
from spinn_utilities.overrides import overrides
from .abstract_sized import AbstractSized, Selector
from .multiple_values_exception import MultipleValuesException
#: :meta private:
R = TypeVar("R")
#: :meta private:
T = TypeVar("T")
#: :meta private:
U = TypeVar("U")
#: :meta private:
IdsType: TypeAlias = Union[Sequence[int], NDArray[numpy.integer]]


def _eq(x: Any, y: Any) -> bool:
    # Lies!
    return numpy.array_equal(numpy.atleast_1d(cast(float, x)),
                             numpy.atleast_1d(cast(float, y)))


def _is_zero(value: Any) -> bool:
    return bool(numpy.isin(0, value))


def is_number(value: T) -> TypeGuard[float]:
    """
    Is the Value a simple integer or float?
    """
    return isinstance(value, Number)


class AbstractList(AbstractSized, Generic[T], metaclass=AbstractBase):
    """
    A ranged implementation of list.

    Functions that change the size of the list are *not* supported.
    These include::

        `__setitem__` where `key` >= `len`
        `__delitem__`
        `append`
        `extend`
        `insert`
        `pop`
        `remove`

    Function that manipulate the list based on values are not supported.
    These include::

        `reverse`, `__reversed__`
        `sort`

    In the current version the IDs are zero-based consecutive numbers so there
    is no difference between value-based IDs and index-based IDs,
    but this could change in the future.

    Supports the following arithmetic operations over the list:

    `+`
        element-wise addition or addition of a single scalar
    `-`
        element-wise subtraction or subtraction of a single scalar
    `*`
        element-wise multiplication or multiplication by a single scalar
    `/`
        element-wise true division or true division by a single scalar
    `//`
        element-wise floor division or floor division by a single scalar
    """
    __slots__ = ("_key", )

    def __init__(self, size: int, key=None):
        """
        :param int size: Fixed length of the list
        :param key: The dict key this list covers.
            This is used only for better Exception messages
        """
        super().__init__(size)
        self._key = key

    @abstractmethod
    def range_based(self) -> bool:
        """
        Shows if the list is suited to deal with ranges or not.

        All list must implement all the range functions,
        but there are times when using ranges will probably be slower than
        using individual values.
        For example the individual values may be stored in a list in which
        case the ranges are created on demand.

        :return: True if and only if Ranged based calls are recommended.
        :rtype: bool
        """
        raise NotImplementedError

    def __len__(self) -> int:
        """
        Size of the list, irrespective of actual values

        :return: the initial and Fixed size of the list
        """
        return self._size

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, AbstractList):
            if self.range_based() and other.range_based():
                return _eq(list(self.iter_ranges()),
                           list(other.iter_ranges()))
        return _eq(list(self), list(other))

    def __ne__(self, other: Any) -> bool:
        if not isinstance(other, AbstractList):
            return True
        return not self.__eq__(other)

    def __str__(self) -> str:
        return str(list(self))

    __repr__ = __str__

    def get_single_value_all(self) -> T:
        """
        If possible, returns a single value shared by the whole list.

        For multiple values use ``for x in list``, ``iter(list)`` or
        ``list.iter``, or one of the ``iter_ranges`` methods

        :return: Value shared by all elements in the list
        :raises ~spinn_utilities.ranged.MultipleValuesException:
            If even one elements has a different value
        """
        # This is not elegant code but as the ranges could be created on the
        # fly the best way.
        only_range: Optional[Tuple[int, int, T]] = None
        for this_range in self.iter_ranges():
            if only_range is not None:
                # If we can get another range, there must be more than one
                # value, so raise the exception
                raise MultipleValuesException(
                    self._key, only_range[2], this_range[2])
            only_range = this_range
        if only_range is None:
            # Pretend we totally failed
            raise StopIteration()
        # There isn't another range, so return the value from the only range
        return only_range[2]

    @abstractmethod
    def get_value_by_id(self, the_id: int) -> T:
        """
        Returns the value for one item in the list.

        :param the_id: One of the IDs of an element in the list
        :type the_id: int
        :return: The value of that element
        """
        raise NotImplementedError

    @abstractmethod
    def get_single_value_by_slice(
            self, slice_start: int, slice_stop: int) -> T:
        """
        If possible, returns a single value shared by the whole slice list.

        For multiple values, use ``for x in list``, ``iter(list)``,
        ``list.iter``, or one of the ``iter_ranges`` methods

        :return: Value shared by all elements in the slice
        :raises ~spinn_utilities.ranged.MultipleValuesException:
            If even one elements has a different value.
            Not thrown if elements outside of the slice have a different value
        """
        raise NotImplementedError

    def __getslice__(self, start: int, stop: int) -> Sequence[T]:
        return list(self.iter_by_slice(start, stop))

    @abstractmethod
    def get_single_value_by_ids(self, ids: IdsType) -> T:
        """
        If possible, returns a single value shared by all the IDs.

        For multiple values, use ``for x in list``, ``iter(list)``,
        ``list.iter``, or one of the ``iter_ranges`` methods.

        :return: Value shared by all elements with these IDs
        :raises ~spinn_utilities.ranged.MultipleValuesException:
            If even one elements has a different value.
            Not thrown if elements outside of the IDs have a different value,
            even if these elements are between the ones pointed to by IDs
        """
        raise NotImplementedError

    def __getitem__(self, selector: Selector) -> Union[Self, T, Sequence[T]]:
        """
        Supports the `list[x]` to return an element or slice of the list.

        :param selector: The int ID, slice
        :return: The element[key] or the slice
        """

        if selector is None:
            return self

        # If the key is a slice, get the values from the slice
        if isinstance(selector, slice):

            # If the slice is continuous, use the continuous slice getter
            if selector.step is None or selector.step == 1:
                return list(self.iter_by_slice(
                    slice_start=selector.start, slice_stop=selector.stop))

            # Otherwise get the items one by one using the start, stop, and
            # step from the slice
            return [
                cast(T, self[i])
                for i in range(*selector.indices(self._size))]

        # If the key is an int, get the single value
        elif self._is_id_type(selector):
            selector = int(cast(int, selector))

            # Handle negative indices
            if selector < 0:
                selector += len(self)
            return self.get_value_by_id(selector)
        else:
            return [self.get_value_by_id(i)
                    for i in cast(Sequence[int], selector)]

    def iter_by_id(self, the_id: int) -> Iterator[T]:
        """
        Fast but *not* update-safe iterator by one ID.

        While ``next`` can only be called once, this is an iterator so it can
        be mixed in with other iterators.

        :param the_id: ID
        :return: yields the elements
        """
        yield self.get_value_by_id(the_id)

    def iter_by_ids(self, ids: IdsType) -> Iterator[T]:
        """
        Fast but *not* update-safe iterator by collection of IDs.

        .. note::
            Duplicate/Repeated elements are yielded for each ID.

        :param ids: IDs
        :return: yields the elements pointed to by IDs
        """
        ranges = self.iter_ranges()
        (_, stop, value) = next(ranges)
        for id_value in ids:

            # If range is too far ahead, reset to start
            if id_value < stop:
                ranges = self.iter_ranges()
                (_, stop, value) = next(ranges)

            # Move on until the ID is in range
            while id_value >= stop:
                (_, stop, value) = next(ranges)

            yield value

    def iter(self) -> Iterator[T]:
        """
        Update-safe iterator of all elements.

        .. note::
            Duplicate/Repeated elements are yielded for each ID

        :return: yields each element one by one
        """
        for id_value in range(self._size):
            yield self.get_value_by_id(id_value)

    def __iter__(self) -> Iterator[T]:
        """
        Fast but *not* update-safe iterator of all elements.

        .. note::
            Duplicate/Repeated elements are yielded for each ID

        :return: yields each element one by one
        """
        try:
            if self.range_based():
                for (start, stop, value) in self.iter_ranges():
                    for _ in range(stop - start):
                        yield value
            else:
                for id_value in range(self._size):
                    yield self.get_value_by_id(id_value)
        except StopIteration:
            return

    def iter_by_slice(self, slice_start: int, slice_stop: int) -> Iterator[T]:
        """
        Fast but *not* update-safe iterator of all elements in the slice.

        .. note::
            Duplicate/Repeated elements are yielded for each ID

        :return: yields each element one by one
        """
        slice_start, slice_stop = self._check_slice_in_range(
            slice_start, slice_stop)
        if self.range_based():
            for (start, stop, value) in \
                    self.iter_ranges_by_slice(slice_start, slice_stop):
                for _ in range(start, stop):
                    yield value
        else:
            for id_value in range(slice_start, slice_stop):
                yield self.get_value_by_id(id_value)

    def iter_by_selector(self, selector: Selector = None) -> Iterator[T]:
        """
        Fast but *not* update-safe iterator of all elements in the slice.

        :param selector: See :py:meth:`AbstractSized.selector_to_ids`
        :return: yields the selected elements one by one
        """
        # No Selector so iter all fast
        if selector is None:
            return iter(self)

        if isinstance(selector, int):
            # Handle negative indices
            if selector < 0:
                selector += len(self)
            return self.iter_by_id(selector)

        # If the key is a slice, get the values from the slice
        if isinstance(selector, slice):

            # If the slice is continuous, use the continuous slice getter
            if selector.step is None or selector.step == 1:
                return self.iter_by_slice(selector.start, selector.stop)
            else:
                (slice_start, slice_stop, step) = selector.indices(self._size)
                return self.iter_by_ids(range(slice_start, slice_stop, step))

        ids = self.selector_to_ids(selector)
        return self.iter_by_ids(ids)

    def get_values(self, selector: Selector = None) -> Sequence[T]:
        """
        Get the value all elements pointed to the selector.

        .. note::
            Unlike ``__get_item__`` this method always returns a list even
            if the selector is a single int.

        :param selector: See :py:meth:`AbstractSized.selector_to_ids`
        :return: returns a list if the item which may be empty or have only
            single value
        :rtype: list
        """
        return list(self.iter_by_selector(selector))

    def __contains__(self, item: T) -> bool:
        return any(_eq(value, item)
                   for (_, _, value) in self.iter_ranges())

    def count(self, x: T) -> int:
        """
        Counts the number of elements in the list with value ``x``.

        :param x:
        :return: count of matching elements
        :rtype: int
        """
        return sum(
            stop - start
            for (start, stop, value) in self.iter_ranges()
            if _eq(value, x))

    def index(self, x: T) -> int:
        """
        Finds the first ID of the first element in the list with the given
        value.

        :param x: The value to find.
        :return: The ID/index
        :raise ValueError: If the value is not found
        """
        for (start, _, value) in self.iter_ranges():
            if _eq(value, x):
                return start
        raise ValueError(f"{x} is not in list")

    @abstractmethod
    def iter_ranges(self) -> Iterator[Tuple[int, int, T]]:
        """
        Fast but *not* update-safe iterator of the ranges.

        :return: yields each range one by one
        """
        raise NotImplementedError

    def iter_ranges_by_id(self, the_id: int) -> Iterator[Tuple[int, int, T]]:
        """
        Iterator of the range for this ID.

        .. note::
            The start and stop of the range will be reduced to just the ID

        This method purpose is so one a control method can select which
        iterator to use.

        :return: yields the one range
        """

        the_id = self._check_id_in_range(the_id)
        for (_, stop, value) in self.iter_ranges():
            if the_id < stop:
                yield (the_id, the_id + 1, value)
                break

    @abstractmethod
    def iter_ranges_by_slice(
            self, slice_start: int, slice_stop: int) -> Iterator[
                Tuple[int, int, T]]:
        """
        Fast but *not* update-safe iterator of the ranges covered by this
        slice.

        .. note::
            The start and stop of the range will be reduced to just the
            IDs inside the slice.

        :return: yields each range one by one
        """
        raise NotImplementedError

    def iter_ranges_by_ids(self, ids: IdsType) -> Iterator[
            Tuple[int, int, T]]:
        """
        Fast but *not* update-safe iterator of the ranges covered by these IDs.

        For consecutive IDs where the elements have the same value a single
        range may be yielded.

        .. note::
            The start and stop of the range will be reduced to just the IDs

        :return: yields each range one by one
        :rtype: iterable
        """
        range_pointer = 0
        result = None
        ranges = list(self.iter_ranges())
        for _id_value in ids:
            id_value = int(_id_value)

            # check if ranges reset so too far ahead
            if id_value < ranges[range_pointer][0]:
                range_pointer = 0
                while id_value > ranges[range_pointer][0]:
                    range_pointer += 1

            # check if pointer needs to move on
            while id_value >= ranges[range_pointer][1]:
                range_pointer += 1
            if result is not None:
                if (result[1] == id_value and _eq(
                        result[2], ranges[range_pointer][2])):
                    result = (result[0], id_value + 1, result[2])
                    continue
                yield result
            result = (id_value, id_value + 1, ranges[range_pointer][2])
        if result is not None:
            yield result

    @abstractmethod
    def get_default(self) -> Optional[T]:
        """
        Gets the default value of the list.
        Just in case we later allow to increase the number of elements.

        :return: Default value
        """
        raise NotImplementedError

    def __add__(self, other: Union[float, AbstractList[float]]
                ) -> AbstractList[float]:
        """
        Support for ``new_list = list1 + list2``.
        Applies the add operator over this and other to create a new list.
        The values of the new list are created on the fly so any changes to
        the original lists are reflected.

        :param other: another list
        :type other: AbstractList
        :return: new list
        :rtype: AbstractList
        :raises TypeError:
        """
        if isinstance(other, AbstractList):
            d_operation: Callable[[Any, float], float] = lambda x, y: x + y
            return DualList(
                left=self, right=other, operation=d_operation)
        if is_number(other):
            s_operation: Callable[[Any], float] = lambda x: x + other
            return SingleList(a_list=self, operation=s_operation)
        raise TypeError("__add__ operation only supported for other "
                        "RangedLists and numerical Values")

    def __sub__(self, other: Union[float, AbstractList[float]]
                ) -> AbstractList[float]:
        """
        Support for ``new_list = list1 - list2``.
        Applies the subtract operator over this and other to create a new list.
        The values of the new list are created on the fly so any changes to
        the original lists are reflected.

        :param other: another list
        :type other: AbstractList
        :return: new list
        :rtype: AbstractList
        :raises TypeError:
        """
        if isinstance(other, AbstractList):
            d_operation: Callable[[Any, float], float] = lambda x, y: x - y
            return DualList(
                left=self, right=other, operation=d_operation)
        if is_number(other):
            s_operation: Callable[[Any], float] = lambda x: x - other
            return SingleList(a_list=self, operation=s_operation)
        raise TypeError("__sub__ operation only supported for other "
                        "RangedLists and numerical Values")

    def __mul__(self, other: Union[float, AbstractList[float]]
                ) -> AbstractList[float]:
        """
        Support for ``new_list = list1 * list2``.
        Applies the multiply operator over this and other.
        The values of the new list are created on the fly so any changes to
        the original lists are reflected.

        :param other: another list
        :type other: AbstractList
        :return: new list
        :rtype: AbstractList
        :raises TypeError:
        """
        if isinstance(other, AbstractList):
            d_operation: Callable[[Any, float], float] = lambda x, y: x * y
            return DualList(
                left=self, right=other, operation=d_operation)
        if is_number(other):
            s_operation: Callable[[Any], float] = lambda x: x * other
            return SingleList(a_list=self, operation=s_operation)
        raise TypeError("__mul__ operation only supported for other "
                        "RangedLists and numerical Values")

    def __truediv__(self, other: Union[float, AbstractList[float]]
                    ) -> AbstractList[float]:
        """
        Support for ``new_list = list1 / list2``.
        Applies the division operator over this and other to create a
        new list.
        The values of the new list are created on the fly so any changes to
        the original lists are reflected.

        :param other: another list
        :type other: AbstractList
        :return: new list
        :rtype: AbstractList
        :raises TypeError:
        """
        if isinstance(other, AbstractList):
            d_operation: Callable[[Any, float], float] = lambda x, y: x / y
            return DualList(
                left=self, right=other, operation=d_operation)
        if is_number(other):
            if _is_zero(other):
                raise ZeroDivisionError()
            s_operation: Callable[[Any], float] = lambda x: x / other
            return SingleList(a_list=self, operation=s_operation)
        raise TypeError("__truediv__ operation only supported for other "
                        "RangedLists and numerical Values")

    def __floordiv__(self, other: Union[float, AbstractList[float]]
                     ) -> AbstractList[int]:
        """
        Support for ``new_list = list1 // list2``.
        Applies the floor division operator over this and other.

        :param other: another list
        :type other: AbstractList
        :return: new list
        :rtype: AbstractList
        :raises TypeError:
        """
        if isinstance(other, AbstractList):
            d_operation: Callable[[Any, float], int] = lambda x, y: int(x // y)
            return DualList(
                left=self, right=other, operation=d_operation)
        if is_number(other):
            if _is_zero(other):
                raise ZeroDivisionError()
            s_operation: Callable[[Any], int] = lambda x: int(x / other)
            return SingleList(a_list=self, operation=s_operation)
        raise TypeError("__floordiv__ operation only supported for other "
                        "RangedLists and numerical Values")

    def apply_operation(
            self, operation: Callable[[T], U]) -> AbstractList[U]:
        """
        Applies a function on the list to create a new one.
        The values of the new list are created on the fly so any changes
        to the original lists are reflected.

        :param operation:
            A function that can be applied over the individual values to
            create new ones.
        :return: new list
        :rtype: AbstractList
        """
        return SingleList(a_list=self, operation=operation)


class SingleList(AbstractList[R], Generic[T, R],
                 metaclass=AbstractBase):
    """
    A List that performs an operation on the elements of another list.
    """
    __slots__ = [
        "_a_list", "_operation"]

    def __init__(self, a_list: AbstractList[T],
                 operation: Callable[[T], R],
                 key: Optional[str] = None):
        """
        :param AbstractList a_list: The list to perform the operation on
        :param callable operation:
            A function which takes a single value and returns the result of
            the operation on that value
        :param key: The dict key this list covers.
            This is used only for better Exception messages
        """
        super().__init__(size=len(a_list), key=key)
        self._a_list = a_list
        self._operation = operation

    @overrides(AbstractList.range_based)
    def range_based(self) -> bool:
        return self._a_list.range_based()

    @overrides(AbstractList.get_value_by_id)
    def get_value_by_id(self, the_id: int) -> R:
        return self._operation(self._a_list.get_value_by_id(the_id))

    @overrides(AbstractList.get_single_value_by_slice)
    def get_single_value_by_slice(
            self, slice_start: int, slice_stop: int) -> R:
        return self._operation(self._a_list.get_single_value_by_slice(
            slice_start, slice_stop))

    @overrides(AbstractList.get_single_value_by_ids)
    def get_single_value_by_ids(self, ids: IdsType) -> R:
        return self._operation(self._a_list.get_single_value_by_ids(ids))

    @overrides(AbstractList.iter_ranges)
    def iter_ranges(self) -> Iterator[Tuple[int, int, R]]:
        for (start, stop, value) in self._a_list.iter_ranges():
            yield (start, stop, self._operation(value))

    @overrides(AbstractList.get_default)
    def get_default(self) -> Optional[R]:
        default = self._a_list.get_default()
        if default is None:
            return None
        return self._operation(default)

    @overrides(AbstractList.iter_ranges_by_slice)
    def iter_ranges_by_slice(
            self, slice_start: int, slice_stop: int
            ) -> Iterator[Tuple[int, int, R]]:
        for (start, stop, value) in \
                self._a_list.iter_ranges_by_slice(slice_start, slice_stop):
            yield (start, stop, self._operation(value))


class DualList(AbstractList[R], Generic[T, U, R],
               metaclass=AbstractBase):
    """
    A list which combines two other lists with an operation.
    """
    __slots__ = [
        "_left", "_operation", "_right"]

    def __init__(self, left: AbstractList[T], right: AbstractList[U],
                 operation: Callable[[T, U], R],
                 key: Optional[str] = None):
        """
        :param AbstractList left: The first list to combine
        :param AbstractList right: The second list to combine
        :param callable operation:
            The operation to perform as a function that takes two values and
            returns the result of the operation
        :param key:
            The dict key this list covers.
            This is used only for better Exception messages
        :raises ValueError: If list are not the same size
        """
        if len(left) != len(right):
            raise ValueError("Two list must have the same size")
        super().__init__(size=len(left), key=key)
        self._left = left
        self._right = right
        self._operation = operation

    @overrides(AbstractList.range_based)
    def range_based(self) -> bool:
        return self._left.range_based() and self._right.range_based()

    @overrides(AbstractList.get_value_by_id)
    def get_value_by_id(self, the_id: int) -> R:
        return self._operation(
            self._left.get_value_by_id(the_id),
            self._right.get_value_by_id(the_id))

    @overrides(AbstractList.get_single_value_by_slice)
    def get_single_value_by_slice(
            self, slice_start: int, slice_stop: int) -> R:
        return self._operation(
            self._left.get_single_value_by_slice(slice_start, slice_stop),
            self._right.get_single_value_by_slice(slice_start, slice_stop))

    @overrides(AbstractList.get_single_value_by_ids)
    def get_single_value_by_ids(self, ids: IdsType) -> R:
        return self._operation(
            self._left.get_single_value_by_ids(ids),
            self._right.get_single_value_by_ids(ids))

    @overrides(AbstractList.iter_by_slice)
    def iter_by_slice(
            self, slice_start: int, slice_stop: int) -> Iterator[R]:
        slice_start, slice_stop = self._check_slice_in_range(
            slice_start, slice_stop)
        if self._left.range_based():
            if self._right.range_based():

                # Both lists are range based
                for (start, stop, value) in \
                        self.iter_ranges_by_slice(slice_start, slice_stop):
                    for _ in range(start, stop):
                        yield value
            else:

                # Left list is range based, right is not
                left_iter = self._left.iter_ranges_by_slice(
                    slice_start, slice_stop)
                right_values = self._right.iter_by_slice(
                    slice_start, slice_stop)
                for (start, stop, left_value) in left_iter:
                    for _ in range(start, stop):
                        yield self._operation(left_value, next(right_values))
        else:
            if self._right.range_based():

                # Right list is range based left is not
                left_values = self._left.iter_by_slice(
                    slice_start, slice_stop)
                right_iter = self._right.iter_ranges_by_slice(
                    slice_start, slice_stop)
                for (start, stop, right_value) in right_iter:
                    for _ in range(start, stop):
                        yield self._operation(next(left_values), right_value)
            else:

                # Neither list is range based
                left_values = self._left.iter_by_slice(slice_start, slice_stop)
                right_values = self._right.iter_by_slice(
                    slice_start, slice_stop)
                while True:
                    try:
                        yield self._operation(
                            next(left_values), next(right_values))
                    except StopIteration:
                        return

    @overrides(AbstractList.iter_ranges)
    def iter_ranges(self) -> Iterator[Tuple[int, int, R]]:
        left_iter = self._left.iter_ranges()
        right_iter = self._right.iter_ranges()
        return self._merge_ranges(left_iter, right_iter)

    @overrides(AbstractList.iter_ranges_by_slice)
    def iter_ranges_by_slice(
            self, slice_start: int, slice_stop: int) -> Iterator[
                Tuple[int, int, R]]:
        left_iter = self._left.iter_ranges_by_slice(slice_start, slice_stop)
        right_iter = self._right.iter_ranges_by_slice(slice_start, slice_stop)
        return self._merge_ranges(left_iter, right_iter)

    def _merge_ranges(self, left_iter: Iterator[Tuple[int, int, T]],
                      right_iter: Iterator[Tuple[int, int, U]]
                      ) -> Iterator[Tuple[int, int, R]]:
        (left_start, left_stop, left_value) = next(left_iter)
        (right_start, right_stop, right_value) = next(right_iter)
        try:
            while True:
                yield (max(left_start, right_start),
                       min(left_stop, right_stop),
                       self._operation(left_value, right_value))
                if left_stop < right_stop:
                    (left_start, left_stop, left_value) = next(left_iter)
                elif left_stop > right_stop:
                    (right_start, right_stop, right_value) = next(right_iter)
                else:
                    (left_start, left_stop, left_value) = next(left_iter)
                    (right_start, right_stop, right_value) = next(right_iter)
        except StopIteration:
            return

    @overrides(AbstractList.get_default)
    def get_default(self) -> Optional[R]:
        l_default = self._left.get_default()
        if l_default is None:
            return None
        r_default = self._right.get_default()
        if r_default is None:
            return None
        return self._operation(l_default, r_default)



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from datetime import timedelta
from time import perf_counter_ns
from typing import Optional
from typing_extensions import Literal

# conversion factor
_NANO_TO_MICRO = 1000.0


class Timer(object):
    """
    A timer used for performance measurements.

    Recommended usage::

        with Timer() as timer:
            ... do stuff that takes time ...

        elapsed_time = timer.measured_interval

    or alternatively::

        timer = Timer()
        timer.start_timing()
        ... do stuff that takes time ...
        elapsed_time = timer.take_sample()

    Mixing these two styles is *not recommended*.
    """

    __slots__ = [

        # The start time when the timer was set off
        "_start_time",

        # The time in the measured section
        "_measured_section_interval"
    ]

    def __init__(self) -> None:
        self._start_time: Optional[int] = None
        self._measured_section_interval: Optional[timedelta] = None

    def start_timing(self) -> None:
        """
        Start the timing. Use :py:meth:`take_sample` to get the end.
        """
        self._start_time = perf_counter_ns()

    def take_sample(self) -> timedelta:
        """
        Describes how long has elapsed since the instance that the
        :py:meth:`start_timing` method was last called.

        :rtype: datetime.timedelta
        """
        time_now = perf_counter_ns()
        diff = time_now - (self._start_time or 0)
        return timedelta(microseconds=diff / _NANO_TO_MICRO)

    def __enter__(self) -> 'Timer':
        self.start_timing()
        return self

    def __exit__(self, *_args) -> Literal[False]:
        self._measured_section_interval = self.take_sample()
        return False

    @property
    def measured_interval(self) -> Optional[timedelta]:
        """
        Get how long elapsed during the measured section.

        :rtype: datetime.timedelta
        """
        return self._measured_section_interval



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations
from typing import (
    Dict, Generic, Iterable, Iterator, Optional, Sequence, Tuple,
    overload, TYPE_CHECKING, Union)
from spinn_utilities.overrides import overrides
from .abstract_dict import AbstractDict, _StrSeq, _Keys
from .abstract_list import IdsType
from .abstract_view import AbstractView, T
if TYPE_CHECKING:
    from .range_dictionary import RangeDictionary


class _IdsView(AbstractView[T], Generic[T]):
    __slots__ = ("_ids", )

    def __init__(self, range_dict: RangeDictionary[T], ids: IdsType):
        """
        Use :py:meth:`RangeDictionary.view_factory` to create views
        """
        super().__init__(range_dict)
        self._ids = tuple(ids)

    def __str__(self) -> str:
        return f"View with IDs: {self._ids}"

    @overrides(AbstractDict.ids)
    def ids(self) -> Sequence[int]:
        return self._ids

    @overload
    def get_value(self, key: str) -> T:
        ...

    @overload
    def get_value(self, key: Optional[_StrSeq]) -> Dict[str, T]:
        ...

    @overrides(AbstractDict.get_value)
    def get_value(self, key: _Keys) -> Union[T, Dict[str, T]]:
        if isinstance(key, str):
            return self._range_dict.get_list(key).get_single_value_by_ids(
                self._ids)
        elif key is None:
            return {
                k: self._range_dict.get_list(k).get_single_value_by_ids(
                    self._ids)
                for k in self._range_dict.keys()}
        else:
            return {
                k: self._range_dict.get_list(k).get_single_value_by_ids(
                    self._ids)
                for k in key}

    @overrides(AbstractDict.set_value)
    def set_value(
            self, key: str, value: T, use_list_as_value: bool = False):
        ranged_list = self._range_dict.get_list(key)
        for _id in self._ids:
            ranged_list.set_value_by_id(the_id=_id, value=value)

    def set_value_by_ids(self, key: str, ids: Iterable[int], value: T):
        """
        Sets a already existing key to the new value. For the view specified.

        :param str key:
        :param iter(int) ids:
        :param value:
        """
        rl = self._range_dict[key]
        for _id in ids:
            rl.set_value_by_id(the_id=_id, value=value)

    @overload
    def iter_all_values(self, key: str, update_safe=False) -> Iterator[T]:
        ...

    @overload
    def iter_all_values(self, key: Optional[_StrSeq],
                        update_safe: bool = False) -> Iterator[Dict[str, T]]:
        ...

    @overrides(AbstractDict.iter_all_values)
    def iter_all_values(self, key: _Keys, update_safe: bool = False):
        if isinstance(key, str):
            yield from self._range_dict.iter_values_by_ids(
                ids=self._ids, key=key, update_safe=update_safe)
        else:
            for _id in self._ids:
                yield self._range_dict.get_values_by_id(key, _id)

    @overload
    def iter_ranges(self, key: str) -> Iterator[Tuple[int, int, T]]:
        ...

    @overload
    def iter_ranges(self, key: Optional[_StrSeq] = None) -> Iterator[Tuple[
            int, int, Dict[str, T]]]:
        ...

    @overrides(AbstractDict.iter_ranges)
    def iter_ranges(self, key: _Keys = None
                    ) -> Union[Iterator[Tuple[int, int, T]],
                               Iterator[Tuple[int, int, Dict[str, T]]]]:
        return self._range_dict.iter_ranges_by_ids(key=key, ids=self._ids)



# Copyright (c) 2021 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class _RequiresSubclassTypeError(TypeError):
    """
    A special exception to handle just the case where the decorator is applied
    so that we can easily figure out when to pass on the extra meta-arguments.

    Do not use outside this file. Outsiders should just see ``TypeError``.
    """


def require_subclass(required_class):
    """
    Decorator that arranges for subclasses of the decorated class to
    require that they are also subclasses of the given class.

    Usage example::

        @require_subclass(AbstractVertex)
        class AbstractVirtual(object):
            ...

    :param type required_class:
        The class that the subclass of the decorated class must be an instance
        of (if that subclass is concrete).
    """

    # Beware! This is all deep shenanigans!
    #
    # The __init_subclass__ stuff is from
    #     https://stackoverflow.com/a/45400374/301832
    # The setattr() call is from:
    #     https://stackoverflow.com/a/533583/301832
    # The classmethod() call is from:
    #     https://stackoverflow.com/a/17930262/301832
    # The use of __class__ to enable super() to work is from:
    #     https://stackoverflow.com/a/43779009/301832
    # The need to do this as a functional decorator is my own discovery;
    # without it, some very weird interactions with meta classes happen and I
    # really don't want to debug that stuff.

    def decorate(target_class):
        # pylint: disable=unused-variable
        __class__ = target_class  # @ReservedAssignment # noqa: F841

        def __init_subclass__(cls, allow_derivation=False, **kwargs):
            if not issubclass(cls, required_class) and not allow_derivation:
                raise _RequiresSubclassTypeError(
                    f"{cls.__name__} must be a subclass "
                    f"of {required_class.__name__} and the derivation was not "
                    "explicitly allowed with allow_derivation=True")
            try:
                super().__init_subclass__(**kwargs)
            except _RequiresSubclassTypeError:
                super().__init_subclass__(
                    allow_derivation=allow_derivation, **kwargs)

        setattr(target_class, '__init_subclass__',
                classmethod(__init_subclass__))
        return target_class
    return decorate



# Copyright (c) 2018 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class _ClassPropertyDescriptor(object):
    """
    A class to handle the management of class properties.
    """

    def __init__(self, fget):
        self.fget = fget

    def __get__(self, obj, klass=None):
        if klass is None:
            klass = type(obj)
        return self.fget.__get__(obj, klass)()


def classproperty(func):
    """
    Defines a property at the class-level.

    Usage::

        class C(object):
            _my_property = "Value"

            @classproperty
            def my_property(cls):
                return cls._my_property
    """
    if not isinstance(func, (classmethod, staticmethod)):
        func = classmethod(func)

    return _ClassPropertyDescriptor(func)



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import defaultdict
from datetime import date
import logging
import math
import os
import sys
from typing import Dict, Iterable, List, TypeVar
from spinn_utilities.config_holder import get_config_bool
from spinn_utilities.log import FormatAdapter
from spinn_utilities.overrides import overrides
from spinn_utilities import logger_utils
import spinn_utilities

logger = FormatAdapter(logging.getLogger(__name__))
#: :meta private:
T = TypeVar("T")


class ProgressBar(object):
    """
    Progress bar for telling the user where a task is up to.
    """
    MAX_LENGTH_IN_CHARS = 60

    TOO_MANY_ERROR = (
        "Too many update steps in progress bar! "
        "This may be a sign that something else has gone wrong!")

    __slots__ = (
        "_number_of_things", "_currently_completed", "_destination",
        "_chars_per_thing", "_chars_done", "_string",
        "_step_character", "_end_character", "_in_bad_terminal",
    )

    def __init__(self, total_number_of_things_to_do,
                 string_describing_what_being_progressed,
                 step_character="=", end_character="|"):
        try:
            self._number_of_things = int(total_number_of_things_to_do)
        except TypeError:

            # Might be dealing with general iterable; better not be infinite
            self._number_of_things = len(list(total_number_of_things_to_do))

        self._currently_completed = 0
        self._chars_per_thing = None
        self._chars_done = 0
        self._string = string_describing_what_being_progressed
        self._destination = sys.stderr
        self._step_character = step_character
        self._end_character = end_character

        # Determine if we are in a "bad" terminal i.e. one that doesn't handle
        # carriage return correctly
        self._in_bad_terminal = "PROGRESS_GOOD_TERMINAL" not in os.environ

        self._create_initial_progress_bar(
            string_describing_what_being_progressed)

    def update(self, amount_to_add=1):
        """
        Update the progress bar by a given amount.

        :param amount_to_add:
        """
        if self._currently_completed + amount_to_add > self._number_of_things:
            logger_utils.error_once(logger, self.TOO_MANY_ERROR)
            return
        self._currently_completed += amount_to_add
        self._check_differences()

    def _print_overwritten_line(self, string: str):
        print("\r" + string, end="", file=self._destination)

    def _print_distance_indicator(self, description: str):
        if description is not None:
            print(description, file=self._destination)

        # Find the mid point
        mid_point = ProgressBar.MAX_LENGTH_IN_CHARS // 2

        # The space between 0% and 50% is the mid-point minus the width of
        # 0% and ~half the width of 50%
        first_space = mid_point - 4

        # The space between 50% and 100% is the mid-point minus the rest of
        # the width of 50% and the width of 100%
        second_space = mid_point - 5

        # Print the progress bar itself
        self._print_distance_line(first_space, second_space)
        if self._in_bad_terminal:
            print("", file=self._destination)
            print(" ", end="", file=self._destination)

    def _print_distance_line(self, first_space, second_space):
        line = f"{self._end_character}0%{' ' * first_space}50%" \
               f"{' ' * second_space}100%{self._end_character}"
        print(line, end="", file=self._destination)

    def _print_progress(self, length: int):
        chars_to_print = length
        if not self._in_bad_terminal:
            self._print_overwritten_line(self._end_character)
        else:
            chars_to_print = length - self._chars_done

        for _ in range(int(chars_to_print)):
            self._print_progress_unit(chars_to_print)
        self._destination.flush()

    def _print_progress_unit(self, chars_to_print):
        # pylint: disable=unused-argument
        print(self._step_character, end='', file=self._destination)

    def _print_progress_done(self) -> None:
        self._print_progress(ProgressBar.MAX_LENGTH_IN_CHARS)
        if not self._in_bad_terminal:
            print(self._end_character, file=self._destination)
        else:
            print("", file=self._destination)

    def _create_initial_progress_bar(self, description):
        if self._number_of_things == 0:
            self._chars_per_thing = ProgressBar.MAX_LENGTH_IN_CHARS
        else:
            self._chars_per_thing = \
                ProgressBar.MAX_LENGTH_IN_CHARS / self._number_of_things
        self._print_distance_indicator(description)
        self._print_progress(0)
        self._check_differences()

    def _check_differences(self):
        expected_chars_done = int(math.floor(
            self._currently_completed * self._chars_per_thing))
        if self._currently_completed == self._number_of_things:
            expected_chars_done = ProgressBar.MAX_LENGTH_IN_CHARS
        self._print_progress(expected_chars_done)
        self._chars_done = expected_chars_done

    def end(self):
        """
        Close the progress bar, updating whatever is left if needed.
        """
        difference = self._number_of_things - self._currently_completed
        self._currently_completed += difference
        self._check_differences()
        self._print_progress_done()

    def __repr__(self):
        return f"<ProgressBar:{self._string}>"

    def __enter__(self):
        """
        Support method to use the progress bar as a context manager::

            with ProgressBar(...) as p:
                ...
                p.update()
                ...
                p.update()
                ...

        This method does not have any parameters because any parameters in the
        with :samp:`ProgressBar(...)` call have been passed to
        :py:meth:`__init__`

        Like :samp:`__new__` this method has to return self as in theory it
        could pass back a different object. Welcome to Python.

        :return: The Progress bar
        """
        return self

    def __exit__(self, exty, exval, traceback):
        self.end()
        return False

    def over(self, collection: Iterable[T],
             finish_at_end: bool = True) -> Iterable[T]:
        """
        Simple wrapper for the cases where the progress bar is being used
        to show progress through the iteration over a single collection.
        The progress bar should have been initialised to the size of the
        collection being iterated over.

        :param ~collections.abc.Iterable collection:
            The base collection (any iterable) being iterated over
        :param bool finish_at_end:
            Flag to say if the bar should finish at the end of the collection
        :return: An iterable. Expected to be directly used in a for.
        :rtype: ~collections.abc.Iterable
        """
        try:
            for item in collection:
                yield item
                self.update()
        finally:
            if finish_at_end:
                self.end()

    def __new__(cls, *args, **kwargs):
        # pylint: disable=unused-argument
        c = cls
        if _EnhancedProgressBar._enabled:
            if get_config_bool("Mode", "I_have_a_sense_of_humour"):
                c = _EnhancedProgressBar
            else:
                _EnhancedProgressBar._enabled = False
        return super().__new__(c)


class _EnhancedProgressBar(ProgressBar):
    """
    Nothing to see here.
    """

    _line_no = 0
    _seq_id = 0
    _step_characters: Dict[int, List[str]] = defaultdict(list)
    _enabled = False
    _DATA_FILE = "progress_bar.txt"

    def _print_progress_unit(self, chars_to_print):
        song_line = self.__line
        if not self._in_bad_terminal:
            print(song_line[0:self._chars_done + chars_to_print],
                  file=self._destination)
        else:
            print(song_line[self._chars_done:self._chars_done + 1],
                  end='', file=self._destination)
            self._chars_done += 1

    def _print_progress_done(self):
        self._print_progress(ProgressBar.MAX_LENGTH_IN_CHARS)
        if not self._in_bad_terminal:
            self._print_overwritten_line(self._end_character)
            for _ in range(ProgressBar.MAX_LENGTH_IN_CHARS):
                print(self._step_character, end='', file=self._destination)
            print(self._end_character, file=self._destination)
        else:
            print("", file=self._destination)
        self.__next_line()

    @property
    def __line(self):
        return _EnhancedProgressBar._step_characters[
            _EnhancedProgressBar._seq_id][_EnhancedProgressBar._line_no]

    @classmethod
    def __next_line(cls):
        if cls._line_no + 1 >= len(cls._step_characters[cls._seq_id]):
            cls._line_no = 0
        else:
            cls._line_no += 1

    @classmethod
    def init_once(cls):
        """
        At startup reads progress bar data from file to be used every time
        """
        cls._enabled = False
        # read in the songs once for performance reasons
        path = os.path.join(
            os.path.dirname(os.path.realpath(spinn_utilities.__file__)),
            cls._DATA_FILE)
        try:
            with open(path, encoding="utf-8") as reader:
                lines = reader.readlines()

            # turn into array of songs, skipping comments and blanks
            for line in lines:
                if line.startswith("#") or line.strip() == "":
                    continue
                bits = line.split(":")
                if len(bits) != 3:
                    # Bad data! Abort!
                    break
                cls._step_characters[bits[0]].append(bits[1])

            # clean up lines so that spaces are still visible
            for _seq_id in cls._step_characters:
                step = cls._step_characters[_seq_id]
                #  pylint: disable=consider-using-enumerate
                for _line_no in range(len(step)):
                    step[_line_no] = step[_line_no].replace(" ", "_")

            # verify that its a special day
            cls._enabled = (
                    date.today().strftime("%m%d") in cls._step_characters)
        except IOError:
            cls._seq_id = 0
        finally:
            cls._line_no = 0
            if cls._enabled:
                cls._seq_id = date.today().strftime("%m%d")
            else:
                # To allow testing on a none special day
                cls._seq_id = "test"


# Perform one-time initialisation
_EnhancedProgressBar.init_once()


class DummyProgressBar(ProgressBar):
    """
    This is a dummy version of the progress bar that just stubs out the
    internal printing operations with code that does nothing. It otherwise
    fails in exactly the same way.
    """
    @overrides(ProgressBar._print_overwritten_line)
    def _print_overwritten_line(self, string: str):
        pass

    @overrides(ProgressBar._print_distance_indicator)
    def _print_distance_indicator(self, description: str):
        pass

    @overrides(ProgressBar._print_progress)
    def _print_progress(self, length: int):
        pass

    @overrides(ProgressBar._print_progress_done)
    def _print_progress_done(self) -> None:
        pass

    def __repr__(self):
        return f"<DummyProgressBar:{self._string}>"


if __name__ == "__main__":  # pragma: no cover
    from time import sleep
    demo = ProgressBar(
        5, "Progress Bar Demonstration", step_character="-", end_character="!")
    for _ in range(5):
        sleep(1)
        demo.update()
    demo.end()
    demo = ProgressBar(30, "Progress Bar Demonstration")
    for _ in range(30):
        sleep(0.1)
        demo.update()
    demo.end()

    _collection = [2, 3, 5, 7, 11, 13, 17]
    demo = ProgressBar(_collection, "Demo over a few primes")
    for prime in demo.over(_collection):
        sleep(0.1)



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations
from collections.abc import Sized
from typing import (
    Any, Callable, Generic, List, Iterable, Iterator, Optional, Sequence,
    Tuple, Union, cast, final)
from typing_extensions import TypeAlias, TypeGuard
from spinn_utilities.overrides import overrides
from spinn_utilities.helpful_functions import is_singleton
from .abstract_sized import Selector
from .abstract_list import AbstractList, T, _eq, IdsType
from .multiple_values_exception import MultipleValuesException

#: The type of a range descriptor
_RangeType: TypeAlias = Tuple[int, int, T]
# The type of things we consider to be a list of values
_ListType: TypeAlias = Union[Callable[[int], T], Sequence[T]]
# The type of value arguments in several places
_ValueType: TypeAlias = Optional[Union[T, _ListType]]


def function_iterator(
        function: Callable[[int], T], size: int,
        ids: Optional[Iterable[int]] = None) -> Iterable[T]:
    """
    Converts a function into an iterator based on size or IDs.
    This so that the function can be used to create a list as in::

        list(function_iterator(lambda x: x * 2 , 3, ids=[2, 4, 6]))

    :param ~collections.abc.Callable[[int],object] function:
        A function with one integer parameter that returns a value
    :param int size:
        The number of elements to put in the list. If used, the function will
        be called with ``range(size)``. Ignored if ``ids`` provided
    :param ~collections.abc.Iterable(int) ids:
        A list of IDs to call the function for or ``None`` to use the size.
    :return: a sequence of values returned by the function
    :rtype: ~collections.abc.Iterable(object)
    """
    if ids is None:
        ids = range(size)
    for _id in ids:
        yield function(_id)


class RangedList(AbstractList[T], Generic[T]):
    """
    A list that is able to efficiently hold large numbers of elements
    that all have the same value.
    """
    __slots__ = [
        "_default", "_ranged_based", "_ranges"]

    def __init__(
            self, size: Optional[int] = None, value: _ValueType = None,
            key=None, use_list_as_value=False):
        """
        :param size:
            Fixed length of the list;
            if ``None``, the value must be a sized object.
        :type size: int or None
        :param value: value to given to all elements in the list
        :type value: object or ~collections.abc.Sized
        :param key: The dict key this list covers.
            This is used only for better Exception messages
        :param bool use_list_as_value: True if the value *is* a list
        """
        if size is None:
            try:
                size = len(value)  # type: ignore[arg-type]
            except TypeError as e:
                raise ValueError("value parameter must have a length to "
                                 "determine the unsupplied size") from e
        super().__init__(size=size, key=key)
        if not use_list_as_value and (
                not self.is_list(value, size) or self.__length(value) != size):
            self._default: Optional[T] = cast(Optional[T], value)
        else:
            self._default = None
        self._ranges: Union[List[T], List[_RangeType]]
        self._ranged_based: Optional[bool] = None
        self.set_value(value, use_list_as_value=use_list_as_value)

    def __length(self, value: Any) -> int:
        if callable(value):
            return self._size
        if not isinstance(value, Sized):
            return 1
        return len(value)

    @overrides(AbstractList.range_based)
    def range_based(self) -> bool:
        return self._ranged_based or False

    @property
    def __the_ranges(self) -> List[_RangeType]:
        assert self._ranged_based
        return cast(List[_RangeType], self._ranges)

    @property
    def __the_values(self) -> List[T]:
        assert not self._ranged_based
        return cast(List[T], self._ranges)

    @overrides(AbstractList.get_value_by_id)
    def get_value_by_id(self, the_id: int) -> T:
        the_id = self._check_id_in_range(the_id)

        # If range based, find the range containing the value and return
        if self._ranged_based:
            for (_, stop, value) in self.__the_ranges:
                if the_id < stop:
                    return value  # pragma: no cover

            # Must never get here because the ID is in range
            raise ValueError  # pragma: no cover

        # Non-range-based so just return the value
        return self.__the_values[the_id]

    @overrides(AbstractList.get_single_value_by_slice)
    def get_single_value_by_slice(
            self, slice_start: int, slice_stop: int) -> T:
        slice_start, slice_stop = self._check_slice_in_range(
            slice_start, slice_stop)

        # If the list is formed of ranges...
        if self._ranged_based:
            found_value = False
            result = None

            # Go through the ranges until in range (assumed sorted)
            for (_, stop, value) in self.__the_ranges:

                # If we have found a range in the slice
                if slice_start < stop:

                    # If we have already found a range in the slice, check the
                    # value is the same
                    if found_value:
                        if not _eq(result, value):
                            raise MultipleValuesException(
                                self._key, result, value)

                    # If this is the first range in the slice, store it
                    else:
                        result = value
                        found_value = True

                    # If we have found a range that finishes outside of the
                    # slice, we have done, and can safely return the value
                    if slice_stop <= stop:
                        return value

            # This must be never possible, as the slices must be in range of
            # the list
            raise ValueError  # pragma: no cover

        # A non-range based list just has lots of single values, so check
        # they are all the same within the slice
        result = self.__the_values[slice_start]
        for value in self.__the_values[slice_start+1: slice_stop]:
            if not _eq(result, value):
                raise MultipleValuesException(self._key, result, value)
        return result

    @overrides(AbstractList.get_single_value_by_ids)
    def get_single_value_by_ids(self, ids: IdsType) -> T:
        # Take the first ID, and then simply check all the others are the same
        # This works for both range-based and non-range-based
        result = self.get_value_by_id(ids[0])
        for id_value in ids[1:]:
            value = self.get_value_by_id(id_value)
            if not _eq(result, value):
                raise MultipleValuesException(self._key, result, value)
        return result

    def __iter__(self) -> Iterator[T]:
        """
        Fast but *not* update-safe iterator of all elements.

        .. note::
            Duplicate/Repeated elements are yielded for each ID.

        :return: yields each element one by one
        """
        if self._ranged_based:
            for (start, stop, value) in self.__the_ranges:
                for _ in range(stop - start):
                    yield value
        else:
            for value in self.__the_values:
                yield value

    @overrides(AbstractList.iter_by_slice)
    def iter_by_slice(self, slice_start: int, slice_stop: int) -> Iterator[T]:
        slice_start, slice_stop = self._check_slice_in_range(
            slice_start, slice_stop)

        # If non-range-based, just go through the values
        if not self._ranged_based:
            yield from self.__the_values[slice_start: slice_stop]
            return

        # Range-based, so go through the ranges that intersect the slice

        # Find the first range within the slice
        ranges = self.iter_ranges()
        (start, stop, value) = next(ranges)
        while stop < slice_start:
            (start, stop, value) = next(ranges)

        # Continue until outside of the slice
        while start < slice_stop:
            first = max(start, slice_start)
            end_point = min(stop, slice_stop)
            for _ in range(end_point - first):
                yield value
            try:
                (start, stop, value) = next(ranges)
            except StopIteration:
                return

    @overrides(AbstractList.iter_ranges)
    def iter_ranges(self) -> Iterator[_RangeType]:
        # If range based just yield the ranges
        if self._ranged_based:
            yield from self.__the_ranges
            return

        # If non-range based, build the ranges
        previous_value = self.__the_values[0]
        previous_start = 0
        current_start = 0
        current_value = None
        any_value = False
        for start, value in enumerate(self.__the_values):
            current_start = start
            current_value = value
            if not _eq(value, previous_value):
                yield (previous_start, start, previous_value)
                previous_start = start
                previous_value = value
            any_value = True
        if any_value:
            yield (previous_start, current_start + 1, cast(T, current_value))

    @overrides(AbstractList.iter_ranges_by_slice)
    def iter_ranges_by_slice(
            self, slice_start: int, slice_stop: int) -> Iterator[_RangeType]:
        slice_start, slice_stop = self._check_slice_in_range(
            slice_start, slice_stop)

        # If range-based, go through ranges that intersect the slice
        if self._ranged_based:
            for (start, stop, value) in self.__the_ranges:
                if slice_start < stop:

                    # The range is updated so that the start and stop values
                    # are within the slice requested
                    yield (max(start, slice_start), min(stop, slice_stop),
                           value)
                    if slice_stop <= stop:
                        break
            return

        # If non-range based, just go through the values
        previous_value = self.__the_values[slice_start]
        previous_start = slice_start
        for index, value in enumerate(
                self.__the_values[slice_start: slice_stop]):
            if not _eq(value, previous_value):
                # Index is one ahead so no need for a + 1 here
                yield (previous_start, slice_start + index, previous_value)
                previous_start = slice_start + index
                previous_value = value
        yield (previous_start, slice_stop, previous_value)

    # pylint: disable=unused-argument
    @final
    def is_list(self, value: _ValueType,
                size: Optional[int]) -> TypeGuard[_ListType]:
        """
        Determines if the value should be treated as a list.

        :param value: The value to examine.
        """
        return self.listness_check(value)

    def listness_check(self, value: _ValueType) -> bool:
        """
        Easier to override in subclasses as not type guard.

        .. note::
            This method can be extended to add other checks for list in which
            case :py:meth:`as_list` must also be extended.

        :param value: The value to examine.
        """
        # Assume any iterable is a list
        if callable(value):
            return True
        return not is_singleton(value)

    def as_list(
            self, value: _ListType, size: int,
            ids: Optional[IdsType] = None) -> List[T]:
        """
        Converts (if required) the value into a list of a given size.
        An exception is raised if value cannot be given size elements.

        .. note::
            This method can be extended to add other conversions to list in
            which case :py:meth:`listness_check` must also be extended.

        :param value:
        :return: value as a list
        :raises Exception: if the number of values and the size do not match
        """
        if callable(value):
            values = list(function_iterator(value, size, ids))
        else:
            values = list(value)
        if len(values) != size:
            raise ValueError(f"The number of values:{len(values)} "
                             f"does not equal the size:{size}")
        return values

    def set_value(self, value: _ValueType, use_list_as_value=False):
        """
        Sets *all* elements in the list to this value.

        .. note::
            Does not change the default.

        :param value: new value
        :param use_list_as_value: True if the value to be set *is* a list
        """

        # If the value to set is a list, just copy the values
        if not use_list_as_value and self.is_list(value, self._size):
            self._ranges = self.as_list(value, self._size)
            self._ranged_based = False

        # Otherwise store the value directly assuming it is the same value
        # for all items
        else:
            self._ranges = [(0, self._size, value)]
            self._ranged_based = True

    def set_value_by_id(self, the_id: int, value: T):
        """
        Sets the value for a single ID to the new value.

        .. note::
            This method only works for a single positive int ID.
            Use ``set`` or ``__set__`` for slices, tuples, lists and negative
            indexes.

        :param int the_id: Single ID
        :param object value: The value to save
        """
        the_id = self._check_id_in_range(the_id)

        # If non-range-based, set the value directly
        if not self._ranged_based:
            self.__the_values[the_id] = value
            return

        # Find the range in which to set the value
        for idx, (start, stop, old_value) in enumerate(self.__the_ranges):
            if the_id < stop:
                break
        else:
            return

        # If already set as needed, do nothing
        if _eq(value, old_value):
            return

        ranges = self.__the_ranges

        # Split the ID out of the range
        ranges[idx] = (the_id, the_id + 1, value)

        # Need a new range after the ID
        if the_id + 1 < stop:
            ranges.insert(idx + 1, (the_id + 1, stop, old_value))

        # Need a new range before the ID
        if the_id > start:
            ranges.insert(idx, (start, the_id, old_value))

        # If not at the last range, update the start and stop value of
        # the next range
        if idx < len(ranges) - 1:
            if _eq(ranges[idx][2], ranges[idx + 1][2]):
                ranges[idx] = (
                    ranges[idx][0], ranges[idx + 1][1], ranges[idx + 1][2])
                ranges.pop(idx + 1)

        # If not at the first range, update the start and stop value
        # of the first range
        if idx > 0:
            if _eq(ranges[idx][2], ranges[idx - 1][2]):
                ranges[idx - 1] = (
                    ranges[idx - 1][0], ranges[idx][1], ranges[idx][2])
                ranges.pop(idx)

    def set_value_by_slice(
            self, slice_start: int, slice_stop: int, value: _ValueType,
            use_list_as_value=False):
        """
        Sets the value for a single range to the new value.

        .. note::
            This method only works for a single positive range.
            Use ``set`` or ``__set__`` for slices, tuples, lists and negative
            indexes.

        :param int slice_start: Start of the range
        :param int slice_stop: Exclusive end of the range
        :param object value: The value to save
        """
        slice_start, slice_stop = self._check_slice_in_range(
            slice_start, slice_stop)
        if slice_start == slice_stop:
            return  # Empty list so do nothing

        # If the value to set is a list, set the values directly
        if not use_list_as_value and self.is_list(
                value, size=slice_stop - slice_start):
            return self._set_values_list(range(slice_start, slice_stop), value)

        # If non-ranged-based, set the values directly
        if not self._ranged_based:
            for id_value in range(slice_start, slice_stop):
                self.__the_values[id_value] = cast(T, value)
            return

        # Skip ranges before start of slice
        ranges = self.__the_ranges
        index = 0
        while index < len(ranges) - 1 and ranges[index][1] <= slice_start:
            index += 1

        # Strip off start of first range in case needed
        (_start, _stop, old_value) = ranges[index]

        # If the slice to set starts after the current range,
        # we may need a new range before the key
        if slice_start > _start:

            # If the values are different, add a new range with the old value
            if not _eq(value, old_value):
                ranges.insert(index, (_start, slice_start, old_value))

                # We have added a value so move on one
                index += 1

                # Change start of old range but leave value for now
                ranges[index] = (slice_start, _stop, old_value)
                (_start, _stop, old_value) = ranges[index]

            # Existing already has this value so start the slice to set
            # at the start of this range
            else:
                slice_start = _start

        # Merge with next if overlaps with that one
        while slice_stop > _stop:
            (_start, _stop, old_value) = ranges.pop(index + 1)
            ranges[index] = (slice_start, _stop, old_value)

        # Split off end of merged range if required
        if slice_stop < _stop:
            ranges[index] = (slice_start, slice_stop, old_value)
            ranges.insert(index+1, (slice_stop, _stop, old_value))

        # merge with previous if same value
        if index > 0 and _eq(ranges[index-1][2], value):
            ranges[index-1] = (ranges[index-1][0], slice_stop, value)
            ranges.pop(index)
            index -= 1

        # merge with next if same value
        if index < len(ranges) - 1 and _eq(ranges[index+1][2], value):
            ranges[index] = (ranges[index][0], ranges[index + 1][1], value)
            ranges.pop(index + 1)

        # set the value in case missed elsewhere
        ranges[index] = (ranges[index][0], ranges[index][1], value)

    def _set_values_list(self, ids: IdsType, value: _ListType):
        values = self.as_list(value=value, size=len(ids), ids=ids)
        for id_value, val in zip(ids, values):
            self.set_value_by_id(id_value, val)

    def set_value_by_ids(
            self, ids: IdsType, value: _ValueType, use_list_as_value=False):
        """
        Sets a already existing key to the new value. For the ids specified.

        :param str key:
        :type ids: iter(int) or numpy.array
        :param value:
        """
        if not use_list_as_value and self.is_list(value, len(ids)):
            self._set_values_list(ids, value)
        else:
            for id_value in ids:
                self.set_value_by_id(id_value, cast(T, value))

    def set_value_by_selector(
            self, selector: Selector, value: _ValueType,
            use_list_as_value=False):
        """
        Support for the ``list[x] =`` format.

        :param selector: A single ID, a slice of IDs or a list of IDs
        :type selector: int or slice or list(int)
        :param object value:
        """
        if selector is None:
            self.set_value(value, use_list_as_value=use_list_as_value)
            return
        if isinstance(selector, slice):
            # Handle a slice
            if selector.step is None or selector.step == 1:
                (start, stop, _) = selector.indices(self._size)
                self.set_value_by_slice(
                    start, stop, value, use_list_as_value=use_list_as_value)
                return

        ids = self.selector_to_ids(selector)
        self.set_value_by_ids(
            ids=ids, value=value, use_list_as_value=use_list_as_value)

    __setitem__ = set_value_by_selector

    def get_ranges(self) -> List[_RangeType]:
        """
        Returns a copy of the list of ranges.

        .. note::
            As this is a copy it will not reflect any updates.

        :rtype: list(tuple(int,int,object))
        """
        if self._ranged_based:
            return list(self.__the_ranges)
        return list(self.iter_ranges())

    def set_default(self, default: Optional[T]):
        """
        Sets the default value.

        .. note::
            Does not change the value of any element in the list.

        :param object default: new default value
        """
        self._default = default

    @overrides(AbstractList.get_default, extend_doc=False)
    def get_default(self) -> Optional[T]:
        """
        Returns the default value for this list.

        :return: Default Value
        :rtype: object
        """
        try:
            return self._default
        except AttributeError as e:
            raise AttributeError("Default value not set.") from e

    def copy_into(self, other: RangedList[T]):
        """
        Turns this List into a of the other list but keep its ID.

        Depth is just enough so that any changes done through the RangedList
        API on other will not change self

        :param RangedList other: Another Ranged List to copy the values from
        """
        # Assume the _default and key remain unchanged
        self._ranged_based = other.range_based()
        # clear the list fast and 2.7 safe
        self._ranges *= 0
        if self._ranged_based:
            self.__the_ranges.extend(other.iter_ranges())
        else:
            self.__the_values.extend(other)

    def copy(self) -> RangedList[T]:
        """
        Creates a copy of this list.

        Depth is just enough so that any changes done through the RangedList
        API on other will not change self

        :return: The copy
        :rtype: RangedList
        """
        clone: RangedList[T] = RangedList(
            self._size, self._default, self._key)
        clone.copy_into(self)
        return clone



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
An implementation of a dictionary and a list that support efficiently working
with ranges of values, used to implement efficient collections for PyNN
population views and assemblies.
"""

from .abstract_dict import AbstractDict
from .abstract_list import AbstractList, DualList, SingleList
from .abstract_sized import AbstractSized
from .abstract_view import AbstractView
from .multiple_values_exception import MultipleValuesException
from .range_dictionary import RangeDictionary
from .ranged_list import RangedList
from .ranged_list_of_lists import RangedListOfList

__all__ = [
    "AbstractDict", "AbstractList", "DualList", "SingleList", "AbstractSized",
    "AbstractView", "MultipleValuesException", "RangeDictionary",
    "RangedList", "RangedListOfList"]



# Copyright (c) 2023 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Types for JSON.
"""

from typing import Dict, List, Union
from typing_extensions import TypeAlias

#: The type of JSON values.
JsonValue: TypeAlias = Union[int, float, str, None, "JsonObject", "JsonArray"]
# NB: The above type is mutually recursive with the definitions below.

#: The type of JSON objects.
JsonObject: TypeAlias = Dict[str, JsonValue]

#: The type of JSON arrays.
JsonArray: TypeAlias = List[JsonValue]

#: The type of JSON arrays of objects. Used for casting.
JsonObjectArray: TypeAlias = List[JsonObject]



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations
from typing import (
    Dict, Generator, Iterable, Iterator, Optional, Sequence, Tuple, Union,
    Generic, overload, TYPE_CHECKING)
from typing_extensions import TypeAlias
from spinn_utilities.overrides import overrides
from .abstract_dict import AbstractDict, T, _StrSeq
from .abstract_sized import AbstractSized
from .abstract_list import IdsType
from .ids_view import _IdsView
from .ranged_list import RangedList
from .single_view import _SingleView
from .slice_view import _SliceView
if TYPE_CHECKING:
    from .abstract_view import AbstractView

_KeyType: TypeAlias = Union[int, slice, Iterable[int]]
_Keys: TypeAlias = Union[None, str, _StrSeq]

_Range: TypeAlias = Tuple[int, int, T]
_SimpleRangeIter: TypeAlias = Iterator[_Range]
_CompoundRangeIter: TypeAlias = Iterator[Tuple[int, int, Dict[str, T]]]


class RangeDictionary(AbstractSized, AbstractDict[T], Generic[T]):
    """
    Main holding class for a range of similar Dictionary object.
    Keys in the dictionary must be str object and can not be removed.
    New keys can be added using the ``dict[str] = value`` format.
    The size (length of the list) is fixed and set at initialisation time.
    """
    __slots__ = [
        "_value_lists"]

    def __init__(self, size: int, defaults: Optional[Dict[str, T]] = None):
        """
        The Object is set up initially where every ID in the range will share
        the same value for each key. All keys must be of type str. The
        default values can be anything, including `None`.

        .. warning::
            Using mutable default values can result in weird problems.

        :param int size: Fixed number of IDs / Length of lists
        :param defaults: Default dictionary where all keys must be str
        :type defaults: dict(str,object)
        """
        super().__init__(size)
        self._value_lists: Dict[str, RangedList[T]] = dict()
        if defaults is not None:
            for key, value in defaults.items():
                self._value_lists[key] = self.list_factory(
                    size=size, value=value, key=key)

    def list_factory(self, size: int, value: T, key: str) -> RangedList[T]:
        """
        Defines which class or subclass of :py:class:`RangedList` to use.

        Main purpose is for subclasses to use a subclass or `RangedList`.
        All parameters are pass through ones to the List constructor

        :param size: Fixed length of the list
        :param value: value to given to all elements in the list
        :param key: The dict key this list covers.
        :return: AbstractList in this case a RangedList
        """
        return RangedList(size, value, key)

    def view_factory(self, key: _KeyType) -> AbstractView:
        """
        Main function for creating views.
        This is the preferred way of creating new views as it checks
        parameters and returns the most efficient view.

        .. note::
            The ``__getitem__`` methods called by `Object[id]` and similar
            defer to this method so are fine to use.

        The ID(s) used are the actual IDs in the range and not indexes on
        the list of IDs

        :param key: A single int ID, a Slice object, or an iterable of int IDs
        :return: A view over the range
        """
        # Key is an int - return single view
        if isinstance(key, int):
            key = self._check_id_in_range(key)
            return _SingleView(range_dict=self, the_id=key)

        # Key is a slice - return a sliced view
        if isinstance(key, slice):
            slice_start, slice_stop = self._check_slice_in_range(
                key.start, key.stop)

            if slice_start >= slice_stop:
                raise KeyError(f"{key} would result in an empty view")

            # Slice is really just one item - return a single view
            if slice_start == slice_stop - 1:
                return _SingleView(range_dict=self, the_id=slice_start)

            # Slice is continuous - return a slice view
            elif key.step is None or key.step == 1:
                return _SliceView(range_dict=self, start=slice_start,
                                  stop=slice_stop)

            # Slice is really a list of integers - change it and continue below
            key = range(self._size)[key]

        # Key can only now be an int iterable so make it a list and check
        key = list(key)
        if not all(isinstance(x, int) for x in key):
            raise KeyError("Only list/tuple of int are supported")

        # Key is really just a single int - return single view
        if len(key) == 1:
            return _SingleView(range_dict=self, the_id=key[0])

        # Key is really just a slice (i.e. one of each key in order without
        # holes) - return a slice view
        if len(key) == key[-1] - key[0] + 1:
            in_order = sorted(key)
            if in_order == key:
                return _SliceView(
                    range_dict=self, start=key[0], stop=key[-1] + 1)

        # Random jumble of int values - return an IDs view
        return _IdsView(range_dict=self, ids=key)

    @overload
    def __getitem__(self, key: str) -> RangedList[T]: ...

    @overload
    def __getitem__(self, key: _KeyType) -> AbstractView: ...

    def __getitem__(self, key):
        """
        Support for the view[x] based the type of the key

        If key is a str, a list type object of :py:class:`AbstractList` is
        returned.
        Otherwise a view (:py:class:`AbstractView`) over part of the IDs in
        the dict is returned.

        Multiple str objects or `None` are not supported as keys here.

        :param key: a str, int, or iterable of int values
        :return: An AbstractList or AbstractView
        :rtype: AbstractList or AbstractView
        """
        if isinstance(key, str):
            return self._value_lists[key]
        return self.view_factory(key=key)

    @overload
    def get_value(self, key: str) -> T: ...

    @overload
    def get_value(self, key: Optional[_StrSeq]) -> Dict[str, T]: ...

    @overrides(AbstractDict.get_value, extend_defaults=True)
    def get_value(self, key: Union[str, None, _StrSeq] = None
                  ) -> Union[T, Dict[str, T]]:
        if isinstance(key, str):
            return self._value_lists[key].get_single_value_all()
        if key is None:
            key = list(self.keys())
        return {
            a_key: self._value_lists[a_key].get_single_value_all()
            for a_key in key}

    @overload
    def get_values_by_id(self, key: str, the_id: int) -> T: ...

    @overload
    def get_values_by_id(
            self, key: Optional[_StrSeq], the_id: int) -> Dict[str, T]: ...

    def get_values_by_id(self, key, the_id) -> Union[T, Dict[str, T]]:
        """
        Same as :py:meth:`get_value` but limited to a single ID.

        :param key: as :py:meth:`get_value`
        :param the_id: single int ID
        :return: See :py:meth:`get_value`
        """
        if isinstance(key, str):
            return self._value_lists[key].get_value_by_id(the_id)
        if key is None:
            key = self.keys()
        return {
            a_key: self._value_lists[a_key].get_value_by_id(the_id)
            for a_key in key}

    def get_list(self, key: str) -> RangedList[T]:
        """
        Gets the storage unit for a single key.

        .. note::
            Mainly intended by Views to access the data for one key directly.

        :param key: a key which must be present in the dict
        :type key: str
        :rtype: :py:class:`.ranged_list.RangedList`
        """
        return self._value_lists[key]

    @overload
    def update_safe_iter_all_values(
            self, key: str, ids: IdsType) -> Generator[T, None, None]: ...

    @overload
    def update_safe_iter_all_values(
            self, key: Optional[_StrSeq],
            ids: IdsType) -> Generator[Dict[str, T], None, None]: ...

    def update_safe_iter_all_values(
            self, key: Union[str, Optional[_StrSeq]],
            ids: IdsType) -> Generator[Union[T, Dict[str, T]], None, None]:
        """
        Same as
        :py:meth:`iter_all_values`
        but limited to a collection of IDs and update-safe.
        """
        return (
            self.get_values_by_id(key=key, the_id=id_value)
            for id_value in ids)

    @overload
    def iter_all_values(
            self, key: str, update_safe: bool = False) -> Iterator[T]:
        ...

    @overload
    def iter_all_values(self, key: Optional[_StrSeq],
                        update_safe: bool = False) -> Iterator[Dict[str, T]]:
        ...

    @overrides(AbstractDict.iter_all_values, extend_defaults=True)
    def iter_all_values(self, key: _Keys, update_safe: bool = False):
        if isinstance(key, str):
            if update_safe:
                return self._value_lists[key].iter()
            return iter(self._value_lists[key])
        else:  # Sub methods will check key type
            if update_safe:
                return self.update_safe_iter_all_values(
                    key, range(self._size))
            return iter(self._values_from_ranges(self.iter_ranges(key)))

    @overload
    def iter_values_by_slice(
            self, slice_start: int, slice_stop: int, key: str,
            update_safe: bool = False) -> Iterator[T]:
        ...

    @overload
    def iter_values_by_slice(
            self, slice_start: int, slice_stop: int,
            key: Optional[_StrSeq] = None,
            update_safe: bool = False) -> Iterator[Dict[str, T]]:
        ...

    def iter_values_by_slice(
            self, slice_start: int, slice_stop: int, key=None,
            update_safe=False):
        """
        Same as :py:meth:`iter_all_values` but limited to a simple slice.
        """
        if isinstance(key, str) and not update_safe:
            return self._value_lists[key].iter_by_slice(
                slice_start=slice_start, slice_stop=slice_stop)
        # Sub methods will check key type
        if update_safe:
            return self.update_safe_iter_all_values(
                key, range(slice_start, slice_stop))
        return self._values_from_ranges(self.iter_ranges_by_slice(
            slice_start=slice_start, slice_stop=slice_stop, key=key))

    @overload
    def iter_values_by_ids(self, ids: IdsType, key: str,
                           update_safe: bool = False) -> Iterator[T]:
        ...

    @overload
    def iter_values_by_ids(
            self, ids: IdsType, key: Optional[_StrSeq] = None,
            update_safe: bool = False) -> Iterator[Dict[str, T]]:
        ...

    def iter_values_by_ids(self, ids: IdsType, key=None, update_safe=False):
        """
        Same as :py:meth:`iter_all_values` but limited to a simple slice.
        """
        if update_safe:
            return self.update_safe_iter_all_values(key, ids)
        return self._values_from_ranges(self.iter_ranges_by_ids(
            key=key, ids=ids))

    @staticmethod
    def _values_from_ranges(ranges: _SimpleRangeIter) -> Iterable[T]:
        for (start, stop, value) in ranges:
            for _ in range(start, stop):
                yield value

    @overrides(AbstractDict.set_value)
    def set_value(
            self, key: str, value: T, use_list_as_value: bool = False):
        self._value_lists[key].set_value(
            value, use_list_as_value=use_list_as_value)

    def __setitem__(self, key: str, value: Union[T, RangedList[T]]):
        """
        Wrapper around set_value to support ``range["key"] =``

        .. note::
            ``range[int] =`` is not supported

        ``value`` can be a single object or ``None`` in
        which case every value in the list is set to that.
        ``value`` can be a collection but
        then it must be exactly the size of all lists in this dictionary.
        ``value`` can be an ``AbstractList``

        :param str key: Existing or *new* dictionary key
        :param value: List or value to create list based on.
        """
        if isinstance(key, str):
            if key in self:
                assert not isinstance(value, RangedList)
                self.set_value(key=key, value=value)
            elif isinstance(value, RangedList):
                assert self._size == len(value)
                self._value_lists[key] = value
                # TODO Make work for other kinds of AbstractList
            else:
                new_list = self.list_factory(
                    size=self._size, value=value, key=key)
                self._value_lists[key] = new_list
        elif isinstance(key, (slice, int, tuple, list)):
            raise KeyError("Setting of a slice/ids not supported")
        else:
            raise KeyError(f"Unexpected key type: {type(key)}")

    @overrides(AbstractDict.ids)
    def ids(self) -> Sequence[int]:
        """
        Returns a list of the IDs in this Range

        :return: a list of the IDs in this Range
        :rtype: list(int)
        """
        return list(range(self._size))

    @overrides(AbstractDict.has_key)
    def has_key(self, key: str) -> bool:
        return key in self._value_lists

    @overrides(AbstractDict.keys)
    def keys(self) -> Iterable[str]:
        return self._value_lists.keys()

    def _merge_ranges(
            self, range_iters: Dict[str, Iterator[Tuple[int, int, T]]]
            ) -> Iterator[Tuple[int, int, Dict[str, T]]]:
        current: Dict[str, T] = dict()
        ranges: Dict[str, Tuple[int, int, T]] = dict()
        start = 0
        stop = self._size
        keys = range_iters.keys()
        for key in keys:
            ranges[key] = next(range_iters[key])
            start = ranges[key][0]
            current[key] = ranges[key][2]
            stop = min(ranges[key][1], stop)
        yield (start, stop, current)
        while stop < self._size:
            current = dict()
            start = self._size
            next_stop = self._size
            for key in keys:
                try:
                    if ranges[key][1] == stop:
                        ranges[key] = next(range_iters[key])
                except StopIteration:
                    return
                start = min(max(ranges[key][0], stop), start)
                next_stop = min(ranges[key][1], next_stop)
                current[key] = ranges[key][2]
            stop = next_stop
            yield (start, stop, current)

    @overload
    def iter_ranges(self, key: str) -> Iterator[Tuple[int, int, T]]:
        ...

    @overload
    def iter_ranges(self, key: Optional[_StrSeq]) -> Iterator[Tuple[
            int, int, Dict[str, T]]]:
        ...

    @overrides(AbstractDict.iter_ranges)
    def iter_ranges(self, key: _Keys = None) -> \
            Union[Iterator[Tuple[int, int, T]],
                  Iterator[Tuple[int, int, Dict[str, T]]]]:
        if isinstance(key, str):
            return self._value_lists[key].iter_ranges()
        if key is None:
            keys = self.keys()
        else:
            keys = key
        return self._merge_ranges({
            a_key: self._value_lists[a_key].iter_ranges()
            for a_key in keys})

    @overload
    def iter_ranges_by_id(
            self, key: str, the_id: Optional[int] = None) -> _SimpleRangeIter:
        ...

    @overload
    def iter_ranges_by_id(
            self, key: Optional[_StrSeq] = None,
            the_id: Optional[int] = None) -> _CompoundRangeIter:
        ...

    def iter_ranges_by_id(self, key=None, the_id=None):
        """
        Same as :py:meth:`iter_ranges` but limited to one ID.

        :param key: see :py:meth:`iter_ranges` parameter key
        :param the_id:
            single ID which is the actual ID and not an index into IDs
        :type the_id: int
        """
        if isinstance(key, str):
            return self._value_lists[key].iter_ranges_by_id(the_id=the_id)
        if key is None:
            key = self.keys()
        return self._merge_ranges({
            a_key: self._value_lists[a_key].iter_ranges_by_id(the_id=the_id)
            for a_key in key})

    @overload
    def iter_ranges_by_slice(
            self, key: str, slice_start: int,
            slice_stop: int) -> _SimpleRangeIter:
        ...

    @overload
    def iter_ranges_by_slice(
            self, key: Optional[_StrSeq], slice_start: int,
            slice_stop: int) -> _CompoundRangeIter:
        ...

    def iter_ranges_by_slice(self, key, slice_start: int, slice_stop: int):
        """
        Same as :py:meth:`iter_ranges` but limited to a simple slice.

        ``slice_start`` and ``slice_stop`` are actual ID values and not
        indexes into the IDs. They must also be actual values, so ``None``,
        ``max_int``, and negative numbers are not supported.

        :param key: see :py:meth:`iter_ranges` parameter ``key``
        :param slice_start: Inclusive i.e. first ID
        :param slice_stop:  Exclusive to last ID + 1
        :return: see :py:meth:`iter_ranges`
        """
        if isinstance(key, str):
            return self._value_lists[key].iter_ranges_by_slice(
                slice_start=slice_start, slice_stop=slice_stop)
        if key is None:
            key = self.keys()
        return self._merge_ranges({
            a_key: self._value_lists[a_key].iter_ranges_by_slice(
                slice_start=slice_start, slice_stop=slice_stop)
            for a_key in key})

    @overload
    def iter_ranges_by_ids(
            self, ids: IdsType, key: str) -> _SimpleRangeIter:
        ...

    @overload
    def iter_ranges_by_ids(
            self, ids: IdsType,
            key: Optional[_StrSeq] = None) -> _CompoundRangeIter:
        ...

    def iter_ranges_by_ids(self, ids: IdsType, key=None):
        """
        Same as :py:meth:`iter_ranges` but limited to a collection of IDs.

        The IDs are actual ID values and not indexes into the IDs

        :param key: see :py:meth:`iter_ranges` parameter ``key``
        :param ids: Collection of IDs in the range
        :return: see :py:meth:`iter_ranges`
        """
        if isinstance(key, str):
            return self._value_lists[key].iter_ranges_by_ids(ids=ids)
        if key is None:
            key = self.keys()
        return self._merge_ranges({
            a_key: self._value_lists[a_key].iter_ranges_by_ids(ids=ids)
            for a_key in key})

    def set_default(self, key: str, default: T):
        """
        Sets the default value for a single key.

        .. note::
            Does not change any values but only changes what ``reset_value``
            would do

        .. warning::
            If called on a view, it sets the default for the *whole* range
            and not just the view.

        :param key: Existing dict key
        :type key: str
        :param default: Value to be used by reset; should not be mutable!
        """
        self._value_lists[key].set_default(default)

    @overrides(AbstractDict.get_default)
    def get_default(self, key: str) -> Optional[T]:
        return self._value_lists[key].get_default()

    def copy_into(self, other: RangeDictionary[T]):
        """
        Turns this dict into a copy of the other dict but keep its id.

        :param RangeDictionary other:
            Another ranged dictionary assumed created by cloning this one
        """
        for key in other.keys():
            value = other[key]
            if isinstance(value, RangedList):
                if key in self:
                    self._value_lists[key].copy_into(value)
                else:
                    self._value_lists[key] = value.copy()
            else:
                self._value_lists[key] = RangedList(
                    len(value), key=key)
                self._value_lists[key].copy_into(value)

    def copy(self) -> RangeDictionary[T]:
        """
        Make a copy of this dictionary. Inner ranged entities are deep copied,
        inner leaf values are shallow copied.

        :return: The copy.
        :rtype: RangeDictionary
        """
        copy: RangeDictionary[T] = RangeDictionary(self._size)
        copy.copy_into(self)
        return copy



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from types import TracebackType
from typing import Optional, Type
from typing_extensions import Literal, Self
from .abstract_base import AbstractBase, abstractmethod


class AbstractContextManager(object, metaclass=AbstractBase):
    """
    Closeable class that supports being used as a simple context manager.
    """

    __slots__ = ()

    @abstractmethod
    def close(self) -> None:
        """
        How to actually close the underlying resources.
        """

    def __enter__(self) -> Self:
        return self

    def __exit__(self, exc_type: Optional[Type], exc_val: Exception,
                 exc_tb: TracebackType) -> Literal[False]:
        self.close()
        return False



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
A trimmed down version of standard Python Abstract Base classes.
"""
from typing import TypeVar
#: :meta private:
T = TypeVar("T")


def abstractmethod(funcobj: T) -> T:
    """
    A decorator indicating abstract methods.

    Requires that the metaclass is :py:class:`AbstractBase` or derived from
    it. A class that has a metaclass derived from :py:class:`AbstractBase`
    cannot be instantiated unless all of its abstract methods are overridden.
    The abstract methods can be called using any of the normal
    'super' call mechanisms.

    Usage::

        class C(object, metaclass=AbstractBase):
            @abstractmethod
            def my_abstract_method(self, ...):
                ...
    """
    funcobj.__isabstractmethod__ = True  # type: ignore[attr-defined]
    return funcobj


class AbstractBase(type):
    """
    Metaclass for defining Abstract Base Classes (AbstractBases).

    Use this metaclass to create an AbstractBase. An AbstractBase can be
    subclassed directly, and then acts as a mix-in class.

    This is a trimmed down version of ABC.
    Unlike ABC you can not register unrelated concrete classes.

    Usage::

        class C(object, metaclass=AbstractBase):
            @abstractmethod
            def my_abstract_method(self, ...):
                ...
    """

    def __new__(mcs, name, bases, namespace, **kwargs):
        # Actually make the class
        abs_cls = super().__new__(mcs, name, bases, namespace, **kwargs)

        # Get set of abstract methods from namespace
        abstracts = set(nm for nm, val in namespace.items()
                        if getattr(val, "__isabstractmethod__", False))

        # Augment with abstract methods from superclasses
        for base in bases:
            for nm in getattr(base, "__abstractmethods__", set()):
                val = getattr(abs_cls, nm, None)
                if getattr(val, "__isabstractmethod__", False):
                    abstracts.add(nm)

        # Lock down the set
        abs_cls.__abstractmethods__ = frozenset(abstracts)
        return abs_cls



# Copyright (c) 2018 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import sys
from .file_converter import FileConverter
from .log_sqllite_database import LogSqlLiteDatabase

ALLOWED_EXTENSIONS = frozenset([".c", ".cpp", ".h"])
SKIPPABLE_FILES = frozenset([
    "common.mk", "Makefile.common",
    "paths.mk", "Makefile.paths",
    "neural_build.mk", "Makefile.neural_build"])


def convert(src, dest, new_dict):
    """
    Converts a whole directory including sub-directories.

    :param str src: Full source directory
    :param str dest: Full destination directory
    :param bool new_dict:
        Whether we should generate a new dictionary/DB.
        If not, we add to the existing one.
    """
    if new_dict:
        LogSqlLiteDatabase(new_dict)

    src_path = os.path.abspath(src)
    if not os.path.exists(src_path):
        raise FileNotFoundError(
            f"Unable to locate source directory {src_path}")
    dest_path = os.path.abspath(dest)
    _convert_dir(src_path, dest_path)


def _convert_dir(src_path, dest_path, make_directories=False):
    """
    Converts a whole directory including sub directories.

    :param str src_path: Full source directory
    :param str dest_path: Full destination directory
    :param bool make_directories: Whether to do `mkdir()` first
    """
    if make_directories:
        _mkdir(dest_path)
    for src_dir, _, file_list in os.walk(src_path):
        dest_dir = os.path.join(dest_path, os.path.relpath(src_dir, src_path))
        if make_directories:
            _mkdir(dest_dir)
        for file_name in file_list:
            _, extension = os.path.splitext(file_name)
            if extension in ALLOWED_EXTENSIONS:
                FileConverter.convert(src_dir, dest_dir, file_name)
            elif file_name in SKIPPABLE_FILES:
                pass
            else:
                source = os.path.join(src_dir, file_name)
                print(f"Unexpected file {source}")


def _mkdir(destination):
    if not os.path.exists(destination):
        os.mkdir(destination)
    if not os.path.exists(destination):
        raise FileNotFoundError(f"mkdir failed {destination}")


if __name__ == '__main__':
    _src = sys.argv[1]
    _dest = sys.argv[2]
    if len(sys.argv) > 3:
        _new_dict = bool(sys.argv[3])
    else:
        _new_dict = False
    convert(_src, _dest, _new_dict)



# Copyright (c) 2021 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations
from tempfile import TemporaryDirectory
from typing import List, Optional

from unittest import SkipTest
from spinn_utilities.exceptions import (
    SpiNNUtilsException,
    SimulatorNotSetupException, SimulatorRunningException,
    SimulatorShutdownException, UnexpectedStateChange)
from spinn_utilities.executable_finder import ExecutableFinder
from .data_status import DataStatus
from .reset_status import ResetStatus
from .run_status import RunStatus
# pylint: disable=protected-access


class _UtilsDataModel(object):
    """
    Singleton data model.

    This class should not be accessed directly please use the DataView and
    DataWriter classes.
    Accessing or editing the data held here directly is *not supported!*

    There may be other DataModel classes which sit next to this one and hold
    additional data. The DataView and DataWriter classes will combine these
    as needed.

    What data is held where and how can change without notice.
    """

    __singleton: Optional[_UtilsDataModel] = None

    __slots__ = [
        "_data_status",
        "_executable_finder",
        "_report_dir_path",
        "_requires_data_generation",
        "_requires_mapping",
        "_reset_status",
        "_run_dir_path",
        "_run_status",
        "_temporary_directory",
    ]

    def __init__(self) -> None:
        self._data_status: DataStatus = DataStatus.NOT_SETUP
        self._executable_finder: ExecutableFinder = ExecutableFinder()
        self._reset_status: ResetStatus = ResetStatus.NOT_SETUP
        self._run_status: RunStatus = RunStatus.NOT_SETUP

    def __new__(cls) -> _UtilsDataModel:
        if cls.__singleton is not None:
            return cls.__singleton
        obj = object.__new__(cls)
        cls.__singleton = obj
        obj._clear()
        return obj

    def _clear(self) -> None:
        """
        Clears out all data.
        """
        self._report_dir_path: Optional[str] = None
        self._hard_reset()

    def _hard_reset(self) -> None:
        """
        Puts all data back into the state expected at graph changed and
        `sim.reset`.
        """
        self._run_dir_path: Optional[str] = None
        self._requires_data_generation = True
        self._requires_mapping = True
        self._temporary_directory: Optional[TemporaryDirectory] = None
        self._soft_reset()

    def _soft_reset(self):
        """
        Puts all data back into the state expected at `sim.reset` but not
        graph changed.
        """
        # Holder for any future values


class UtilsDataView(object):
    """
    A read only view of the data available at each level.

    .. note::
        The state model of this class is designed primarily to support
        sPyNNaker.

    All methods are class methods so can be accessed directly without
    instantiating a view.
    There is a stack of subclasses such as `MachineDataView`, `FecDataView`,
    `SpynnakerDataView` (and more). All inherit all methods.
    We reserve the right to override methods defined in one View in subclasses.

    There are also Writer classes which override the Views but these are
    specifically designed to only be usable in unit tests and by the simulator
    (ASB) directly.

    You should use the data view most appropriate to what you are doing i.e.
    If you are accessing it from a class or function in FEC,
    use `FecDataView` but if you are accessing it from sPyNNaker,
    use `SpynnakerDataView`.
    This will ensure that any changes to the view local to the code will
    affect all code in that package

    The objects accessed this way should not be changed or added to.
    Changing or adding to any object accessed is unsupported as bypasses any
    check or updates done in the writer(s).
    Objects returned could be changed to immutable versions without notice!

    The `get...` methods will either return a valid value or
    raise an exception if the data is currently not available.
    The `get` methods will will not return `None` unless specifically
    documented to do so.
    As a reasonable effort is made the setters to verify the data types,
    the get methods can be expected to return the correct type.

    There are also several semantic sugar `get...` methods.
    Some are slightly faster but many are just to make the code more readable.
    Some semantic sugar methods do not start with get to keep the same name as
    the existing function on the object has.

    The `iterate...` methods offer a view over the collections within
    mutable data objects, particularly ones changed between runs.
    There is no guarantee if the returned iterator will or will not reflect
    any changes to the underlying data object,
    nor that how a method behaves in this way does not change over time.
    So the methods should be called for every iteration.

    Each `iterate...` method will have a corresponding `get_n...` which you
    need to do instead of `len(iterate...)` as we reserve the right to make
    any `iterate...` method return an iterable which does not support `len`
    without notice.

    `add...` methods allow for the scripts directly or indirectly to add extra
    values.
    They allow the view to add extra safety such as type checking.
    They will raise an exception if called while the simulator is running.

    The `has...` methods will return True if the value is known and False if
    not.
    Semantically they are the same as checking if the get raises an exception.
    They may be faster if the object needs to be generated on the fly or
    protected to be made immutable.
    `has...` methods have been added where found needed.
    More can easily be added if required.

    The `is...` methods will return a Boolean value to say the simulator is in
    the expected state.
    They may throw an exception if called at an unexpected time.
    For example if called before `sim.setup` or after `sim.end`.

    While how and where the underpinning DataModel(s) store data can change
    without notice, methods in View classes can be considered a supported API.
    """

    __data = _UtilsDataModel()
    __slots__ = ()

    @classmethod
    def _exception(cls, data: str) -> SpiNNUtilsException:
        """
        The most suitable no data Exception based on the status.

        :param str data: Name of the data not found
        :rtype: ~spinn_utilities.exceptions.SpiNNUtilsException
        """
        return cls.__data._data_status.exception(data)

    # Status checks

    @classmethod
    def _is_mocked(cls) -> bool:
        """
        Checks if the view is in mocked state.

        :rtype: bool
        """
        return cls.__data._data_status == DataStatus.MOCKED

    @classmethod
    def is_hard_reset(cls) -> bool:
        """
        Check if the system has been hard reset since the last run
        *finished*.

        .. warning::
            During the first run after reset this continues to return True!

        Returns False after a reset that was considered soft.

        :rtype: bool
        """
        return cls.__data._reset_status == ResetStatus.HARD_RESET

    @classmethod
    def is_soft_reset(cls) -> bool:
        """
        Check if the system has been soft reset since the last run *finished*.

        .. warning::
            During the first run after reset this continues to return True!

        Returns False after a reset that was considered hard.

        :rtype: bool
        """
        return cls.__data._reset_status == ResetStatus.SOFT_RESET

    @classmethod
    def is_ran_ever(cls) -> bool:
        """
        Check if the simulation has run at least once, ignoring resets.

        :rtype: bool
        :raises NotImplementedError:
            If this is called from an unexpected state
        """
        if cls.__data._reset_status == ResetStatus.SETUP:
            return False
        if cls.__data._reset_status in [
                ResetStatus.HAS_RUN, ResetStatus.SOFT_RESET,
                ResetStatus.HARD_RESET]:
            return True
        raise NotImplementedError(
            f"This call was not expected with reset status "
            f"{cls.__data._reset_status}")

    @classmethod
    def is_ran_last(cls) -> bool:
        """
        Checks if the simulation has run and not been reset.

        :rtype: bool
        :raises NotImplementedError:
            If this is called from an unexpected state
        """
        if cls.__data._reset_status == ResetStatus.HAS_RUN:
            return True
        if cls.__data._reset_status in [
                ResetStatus.SETUP, ResetStatus.SOFT_RESET,
                ResetStatus.HARD_RESET]:
            return False
        raise NotImplementedError(
            f"This call was not expected with reset status "
            f"{cls.__data._reset_status}")

    @classmethod
    def is_reset_last(cls) -> bool:
        """
        Reports if `sim.reset` called since the last `sim.run`.

        Unlike :py:meth:`is_soft_reset` and :py:meth:`is_hard_reset` this
        method return False during any `sim.run`.

        It also returns False after a `sim.stop` or `sim.end` call starts

        :rtype: bool
        :raises NotImplementedError:
            If this is called from an unexpected state
        """
        if cls.__data._reset_status in [
                ResetStatus.SETUP, ResetStatus.HAS_RUN]:
            return False
        if cls.__data._reset_status in [
                ResetStatus.SOFT_RESET, ResetStatus.HARD_RESET]:
            if cls.__data._run_status == RunStatus.NOT_RUNNING:
                return True
            if cls.__data._run_status == RunStatus.IN_RUN:
                return False
        if cls.__data._run_status in [
                RunStatus.STOP_REQUESTED, RunStatus.STOPPING,
                RunStatus.SHUTDOWN]:
            raise SimulatorShutdownException(
                "This call is not supported after sim.stop/end or "
                "sim.end has been called")
        raise NotImplementedError(
            f"This call was not expected with reset status "
            f"{cls.__data._reset_status} and run status "
            f"{cls.__data._run_status}")

    @classmethod
    def is_no_stop_requested(cls) -> bool:
        """
        Checks that a stop request has not been sent.

        :rtype: bool
        :raises NotImplementedError:
            If this is called from an unexpected state
        """
        if cls.__data._run_status == RunStatus.IN_RUN:
            return True
        if cls.__data._run_status == RunStatus.STOP_REQUESTED:
            return False
        raise NotImplementedError(
            f"This call was not expected with run status "
            f"{cls.__data._run_status}")

    @classmethod
    def is_running(cls) -> bool:
        """
        Checks if there is currently a simulation running.

        That is a call to run has started but not yet stopped.

        :rtype: bool
        """
        return cls.__data._run_status in [
            RunStatus.IN_RUN, RunStatus.STOP_REQUESTED]

    @classmethod
    def check_valid_simulator(cls) -> None:
        """
        Throws an error if there is no simulator.

        :raises SimulatorNotSetupException: If called before `sim.setup`
        :raises SimulatorShutdownException: If called after `sim.end`
        """
        if cls.__data._run_status in [
                RunStatus.NOT_RUNNING, RunStatus.IN_RUN,
                RunStatus.STOP_REQUESTED, RunStatus.STOPPING]:
            return
        if cls.__data._data_status == DataStatus.MOCKED:
            return
        if cls.__data._run_status == RunStatus.NOT_SETUP:
            raise SimulatorNotSetupException(
                "This call is not supported before setup has been called")
        if cls.__data._run_status == RunStatus.SHUTDOWN:
            raise SimulatorShutdownException(
                "This call is not valid after end or stop has been called")
        raise NotImplementedError(
            f"Unexpected run state: {cls.__data._run_status}")

    @classmethod
    def check_user_can_act(cls) -> None:
        """
        Checks if the status is such that users can be making calls.

        This does *not* error in the Mocked state

        :raises SimulatorRunningException: If `sim.run` is currently running
        :raises SimulatorNotSetupException: If called before `sim.setup`
        :raises SimulatorShutdownException: If called after `sim.end`
        """
        if cls.__data._run_status == RunStatus.NOT_RUNNING:
            return
        if cls.__data._data_status == DataStatus.MOCKED:
            return
        if cls.__data._run_status in [
                RunStatus.IN_RUN, RunStatus.STOPPING,
                RunStatus.STOP_REQUESTED]:
            raise SimulatorRunningException(
                f"This call is not supported while the simulator is running "
                f"and in state {cls.__data._run_status}")
        cls.check_valid_simulator()

    @classmethod
    def is_setup(cls) -> bool:
        """
        Checks to see if there is already a simulator.

        :rtype: bool
        :raises NotImplementedError:
            If this is called from an unexpected state
        """
        if cls.__data._run_status in [RunStatus.NOT_SETUP, RunStatus.SHUTDOWN]:
            return False
        if cls.__data._run_status in [
                RunStatus.NOT_RUNNING, RunStatus.IN_RUN,
                RunStatus.STOP_REQUESTED, RunStatus.STOPPING]:
            return True
        raise NotImplementedError(
            f"This call was not expected with run status "
            f"{cls.__data._run_status}")

    @classmethod
    def is_user_mode(cls) -> bool:
        """
        Determines if simulator is currently not running so user is
        accessing data.

        This returns False in the Mocked state.

        :rtype: bool
        :raises NotImplementedError:
            If the data has not yet been set up or on an unexpected run_status
        """
        if cls.__data._run_status in [
                RunStatus.IN_RUN, RunStatus.STOPPING,
                RunStatus.STOP_REQUESTED]:
            return False
        if cls.__data._run_status in [
                RunStatus.NOT_RUNNING, RunStatus.SHUTDOWN]:
            return True
        if cls._is_mocked():
            return False
        raise NotImplementedError(
            f"Unexpected with RunStatus {cls.__data._run_status}")

    @classmethod
    def is_stop_already_requested(cls) -> bool:
        """
        Checks if there has already been a request stop.

        Also checks the state is such that a stop request makes sense.

        :return: True if the stop has already been requested
            or if the system is stopping or has already stopped
            False if the stop request makes sense.
        :rtype: bool
        :raises NotImplementedError:
            If this is called from an unexpected state
        :raises SpiNNUtilsException:
            If the stop_run was not expected in the current state.
        """
        if cls.__data._run_status == RunStatus.IN_RUN:
            return False
        if cls.__data._run_status == RunStatus.STOP_REQUESTED:
            return True
        cls.check_valid_simulator()
        if cls.is_ran_ever():
            raise UnexpectedStateChange(
                "Calling stop after run has finished does not make sense")
        raise UnexpectedStateChange(
                "Calling stop before run does not make sense")

    @classmethod
    def is_shutdown(cls) -> bool:
        """
        Determines if simulator has already been shutdown.

        This returns False in the Mocked state

        :rtype: bool
        """
        return cls.__data._run_status == RunStatus.SHUTDOWN

    # Report directories
    # Remainder in FecDataView

    @classmethod
    def _temporary_dir_path(cls) -> str:
        """
        The path to an existing temporary directory, creating it if needed.

        :raises ~spinn_utilities.exceptions.SpiNNUtilsException:
            if not in Mocked state
        """
        if cls.__data._temporary_directory is None:
            cls.__data._temporary_directory = TemporaryDirectory()
        return cls.__data._temporary_directory.name

    @classmethod
    def get_run_dir_path(cls) -> str:
        """
        Returns the path to the directory that holds all the reports for run.

        This will be the path used by the last run call or to be used by
        the next run if it has not yet been called.

        .. note::
            In unit test mode this returns a temporary directory
            shared by all path methods.

        :rtype: str
        :raises ~spinn_utilities.exceptions.SpiNNUtilsException:
            If the run_dir_path is currently unavailable
        """
        if cls.__data._run_dir_path:
            return cls.__data._run_dir_path
        if cls._is_mocked():
            return cls._temporary_dir_path()
        raise cls._exception("run_dir_path")

    @classmethod
    def get_executable_finder(cls) -> ExecutableFinder:
        """
        The ExcutableFinder object created at time code is imported.

        :rtype: ExcutableFinder
        """
        return cls.__data._executable_finder

    @classmethod
    def register_binary_search_path(cls, search_path: str):
        """
        Register an additional binary search path for executables.

        Syntactic sugar for `get_executable_finder().add_path()`

        :param str search_path: absolute search path for binaries
        """
        cls.__data._executable_finder.add_path(search_path)

    @classmethod
    def get_executable_path(cls, executable_name: str) -> str:
        """
        Finds an executable within the set of folders. The set of folders
        is searched sequentially and the first match is returned.

        Syntactic sugar for `get_executable_finder().get_executable_path()`

        :param str executable_name: The name of the executable to find
        :return: The full path of the discovered executable
        :rtype: str
        :raises KeyError: If no executable was found in the set of folders
        """
        return cls.__data._executable_finder.get_executable_path(
            executable_name)

    @classmethod
    def get_executable_paths(cls, executable_names: str) -> List[str]:
        """
        Finds each executables within the set of folders.

        The names are assumed to be comma separated
        The set of folders is searched sequentially
        and the first match for each name is returned.

        Names not found are ignored and not added to the list.

        Syntactic sugar for `get_executable_finder().get_executable_paths()`

        :param str executable_names: The name of the executable to find.
            Assumed to be comma separated.
        :return:
            The full path of the discovered executable, or ``None`` if no
            executable was found in the set of folders
        :rtype: list(str)
        """
        return cls.__data._executable_finder.get_executable_paths(
            executable_names)

    @classmethod
    def get_requires_data_generation(cls) -> bool:
        """
        Reports if data generation is required.

        Set to True at the start and by any change that could require
        data generation or mapping
        Remains True during the first run after a data change
        Only set to False at the *end* of the first run

        :rtype: bool
        """
        return cls.__data._requires_data_generation

    @classmethod
    def set_requires_data_generation(cls) -> None:
        """
        Sets `requires_data_generation` to True.

        Only the end of a run can set it to False
        """
        cls.check_user_can_act()
        cls.__data._requires_data_generation = True

    @classmethod
    def get_requires_mapping(cls) -> bool:
        """
        Reports if mapping is required.

        Set to True at the start and by any change that could require
        any mapping stage to be called
        Remains True during the first run after a requires mapping.
        Only set to False at the *end* of the first run

        :rtype: bool
        """
        return cls.__data._requires_mapping

    @classmethod
    def set_requires_mapping(cls) -> None:
        """
        Sets `requires_mapping` and `requires_data_generation` to True.

        Only the end of a run can set it to False
        """
        cls.check_user_can_act()
        cls.__data._requires_mapping = True
        cls.__data._requires_data_generation = True

    @classmethod
    def _mock_has_run(cls) -> None:
        """
        Mock the status as if run has been called and finished!

        .. warning::

            *ONLY FOR USE IN UNITTESTS*

            Any use outside of unit tests is *not* supported and will cause
            errors!
        """
        cls.__data._run_status = RunStatus.NOT_RUNNING
        cls.__data._reset_status = ResetStatus.HAS_RUN
        cls.__data._requires_data_generation = False
        cls.__data._requires_mapping = False

    @classmethod
    def raise_skiptest(cls, reason: str,
                       parent: Optional[Exception] = None) -> None:
        """
        Sets the status as shutdown and raises a SkipTest

        :param str reason: Message for the Skip
        :param parent: Exception which triggered the skip if any
        :type parent: Exception or None
        :raises: SkipTest very time called
        """
        cls.__data._data_status = DataStatus.SHUTDOWN
        cls.__data._run_status = RunStatus.SHUTDOWN
        if parent is None:
            raise SkipTest(reason)
        else:
            raise SkipTest(reason) from parent



# Copyright (c) 2022 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import datetime
from typing import List, Optional, Tuple
from .abstract_base import abstractmethod


class LogStore(object):
    """
    API supported by classes that can store logs for later retrieval.
    """

    @abstractmethod
    def store_log(self, level: int, message: str,
                  timestamp: Optional[datetime] = None):
        """
        Writes the log message for later retrieval.

        :param int level: The logging level.
        :param str message: The logged message.
        :param timestamp: The time-stamp of the message.
        :type timestamp: ~datetime.datetime or None
        """
        raise NotImplementedError

    @abstractmethod
    def retreive_log_messages(
            self, min_level: int = 0) -> List[Tuple[int, str]]:
        """
        Retrieves all log messages at or above the `min_level`.

        :param int min_level:
            Constraint on the minimum logging level to retrieve.
        :return:
            A list of messages that satisfy the constraint.
        :rtype: list(tuple(int, str))
        """
        raise NotImplementedError

    @abstractmethod
    def get_location(self) -> str:
        """
        Retrieves the location of the log store.

        :rtype: str
        """
        raise NotImplementedError



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

_already_issued = set()


def warn_once(logger, msg):
    """
    Write a warning message to the given logger where that message should only
    be written to the logger once.

    :param logger: The logger to write to.
    :param msg: The message to write.

        .. note::
            This must be a substitution-free message.
    """
    if msg in _already_issued:
        return
    _already_issued.add(msg)
    logger.warning(msg)


def error_once(logger, msg):
    """
    Write an error message to the given logger where that message should only
    be written to the logger once.

    :param logger: The logger to write to.
    :param msg: The message to write.

        .. note::
            This must be a substitution-free message.
    """
    if msg in _already_issued:
        return
    _already_issued.add(msg)
    logger.error(msg)


def reset():
    """
    Clear the store of what messages have already been seen.

    .. note::
        Only normally needed by test code.
    """
    global _already_issued     # pylint: disable=global-statement
    _already_issued = set()
